# index.py
''' Content indexer '''

import concurrent.futures
import logging
import os
import threading
import time
import typing

import watchdog.events
import watchdog.observers
from pony import orm

from . import category, entry, model, utils

LOGGER = logging.getLogger(__name__)

ENTRY_TYPES = ['.md', '.htm', '.html']
CATEGORY_TYPES = ['.cat', '.meta']


class Indexer:
    """ Class which handles the scheduling of file indexing """
    # pylint:disable=too-many-instance-attributes
    QUEUE_ITEM = typing.Tuple[str, typing.Optional[str], int]

    def __init__(self, app, wait_time: float):
        self._app = app
        # pylint:disable=consider-using-with
        self._thread_pool = concurrent.futures.ThreadPoolExecutor(
            max_workers=1,
            thread_name_prefix="Indexer")
        self._pending: typing.Set[Indexer.QUEUE_ITEM] = set()
        self._wait_count = 0
        self._in_progress = 0
        self._lock = threading.Lock()
        self._count_lock = threading.Lock()
        self._running: typing.Optional[concurrent.futures.Future] = None
        self._wait_time = wait_time
        self.last_indexed: typing.Optional[str] = None

    @property
    def in_progress(self) -> bool:
        """ Returns whether there's an active or pending indexer task """
        with self._lock:
            return self._wait_count > 0

    @property
    def queue_size(self) -> int:
        """ Returns the number of queued work items """
        total = 0
        with self._count_lock:
            total += self._wait_count
        with self._lock:
            total += self._in_progress
        return total

    def scan_file(self,
                  fullpath: str, relpath: typing.Optional[str],
                  fixup_pass: int,
                  wait: bool = False):
        """ Scan a file for the index

        fullpath -- The full path to the file
        relpath -- The path to the file, relative to its base directory
        fixup_pass -- Which phase of the fixup process we're on
        wait -- whether to wait for the initial safety delay

        This calls into various modules' scanner functions; the expectation is that
        the scan_file function will return a truthy value if it was scanned
        successfully, False if it failed, and None if there is nothing to scan.
        """

        with self._lock:
            self._pending.add((fullpath, relpath, fixup_pass))
            self._in_progress += 1
        self._start_scan(self._wait_time if wait else 0)

    def submit(self, func, *args, **kwargs):
        """ Schedule a task into the task pool """
        with self._count_lock:
            self._wait_count += 1

        def do_task():
            try:
                with self._app.app_context():
                    func(*args, **kwargs)
            except Exception:  # pylint:disable=broad-except
                LOGGER.exception("Task failed")
            with self._count_lock:
                self._wait_count -= 1

        return self._thread_pool.submit(do_task)

    def _start_scan(self, wait: float = 0):
        with self._lock:
            if not self._running or self._running.done():
                if wait:
                    # busywait the worker thread to wait for things to settle down
                    # (and to batch up a bunch of updates in a row)
                    self.submit(time.sleep, wait)

                # run the actual pending task
                self._running = self.submit(self._scan_pending)

    def _scan_pending(self):
        """ Scan all the pending items """
        with self._lock:
            items = self._pending
            self._in_progress = len(items)
            self._pending = set()
        LOGGER.debug("Processing %d files", len(items))

        # process the known items
        for item in items:
            self._scan_file(*item)
            with self._lock:
                self._in_progress -= 1

        # and then schedule a catchup for anything that happened
        # while this scan was happening
        if items:
            self.submit(self._start_scan)
        else:
            with self._lock:
                self._in_progress = 0

    def _scan_file(self, fullpath: str, relpath: typing.Optional[str], fixup_pass: int):
        LOGGER.debug("Scanning file: %s (%s) pass=%d", fullpath, relpath, fixup_pass)

        def do_scan() -> typing.Optional[bool]:
            """ helper function to do the scan and gather the result """
            _, ext = os.path.splitext(fullpath)

            try:
                if ext in ENTRY_TYPES:
                    LOGGER.info("Scanning entry: %s", fullpath)
                    return entry.scan_file(fullpath, relpath, fixup_pass)

                if ext in CATEGORY_TYPES:
                    LOGGER.info("Scanning meta info: %s", fullpath)
                    return category.scan_file(fullpath, relpath)

                return None
            except Exception:  # pylint: disable=broad-except
                LOGGER.exception("Got error parsing %s", fullpath)
                return False

        result = do_scan()
        if result is False and fixup_pass < 5:
            LOGGER.info("Scheduling fixup pass %d for %s", fixup_pass + 1, fullpath)
            self.scan_file(fullpath, relpath, fixup_pass + 1)
        else:
            LOGGER.debug("%s complete", fullpath)
            set_fingerprint(fullpath)
            self.last_indexed = fullpath
        return result


def last_indexed() -> typing.Optional[str]:
    """ information about the most recently indexed file, for cache-busting
    purposes """
    from .flask_wrapper import current_app
    ref = current_app.indexer.last_indexed
    if ref:
        return get_last_fingerprint(ref)
    return None


def queue_size() -> typing.Optional[int]:
    """ Return the approximate length of the work queue """
    from .flask_wrapper import current_app
    return current_app.indexer.queue_size


def in_progress() -> bool:
    """ Return if there's an index in progress """
    from .flask_wrapper import current_app
    return current_app.indexer.in_progress


def is_scannable(fullpath) -> bool:
    """ Determine if a file needs to be scanned """
    _, ext = os.path.splitext(fullpath)
    return ext in ENTRY_TYPES or ext in CATEGORY_TYPES


@orm.db_session
def get_last_fingerprint(fullpath: str) -> typing.Optional[str]:
    """ Get the last known fingerprint for a file """
    record = model.FileFingerprint.get(file_path=fullpath)
    if record:
        return record.fingerprint
    return None


@orm.db_session(retry=5)
def set_fingerprint(fullpath, fingerprint=None):
    """ Set the last known modification time for a file """
    try:
        fingerprint = fingerprint or utils.file_fingerprint(fullpath)

        record = model.FileFingerprint.get(file_path=fullpath)
        if record and record.fingerprint != fingerprint:
            record.set(fingerprint=fingerprint,
                       file_mtime=os.stat(fullpath).st_mtime)
        elif not record:
            record = model.FileFingerprint(
                file_path=fullpath,
                fingerprint=fingerprint,
                file_mtime=os.stat(fullpath).st_mtime)
        orm.commit()
    except FileNotFoundError:
        orm.delete(fp for fp in model.FileFingerprint if fp.file_path == fullpath)


class IndexWatchdog(watchdog.events.PatternMatchingEventHandler):
    """ Watchdog handler """

    def __init__(self, indexer, content_dir):
        super().__init__(ignore_directories=True)
        self._indexer = indexer
        self.content_dir = content_dir

    def update_file(self, fullpath):
        """ Update a file """
        LOGGER.debug("Scheduling reindex of %s", fullpath)
        relpath = os.path.relpath(fullpath, self.content_dir)
        self._indexer.scan_file(fullpath, relpath, False)

    def on_created(self, event):
        """ on_created handler """
        LOGGER.debug("file created: %s", event.src_path)
        if not event.is_directory:
            self.update_file(event.src_path)

    def on_modified(self, event):
        """ on_modified handler """
        LOGGER.debug("file modified: %s", event.src_path)
        if not event.is_directory:
            self.update_file(event.src_path)

    def on_moved(self, event):
        """ on_moved handler """
        LOGGER.debug("file moved: %s -> %s", event.src_path, event.dest_path)
        if not event.is_directory:
            self.update_file(event.src_path)
            self.update_file(event.dest_path)

    def on_deleted(self, event):
        """ on_deleted handler """
        LOGGER.debug("File deleted: %s", event.src_path)
        if not event.is_directory:
            self.update_file(event.src_path)


def background_scan(content_dir):
    """ Start background scanning a directory for changes """
    from .flask_wrapper import current_app
    observer = watchdog.observers.Observer()
    observer.schedule(IndexWatchdog(current_app.indexer, content_dir),
                      content_dir, recursive=True)
    logging.info("Watching %s for changes", content_dir)
    observer.start()


def prune_missing(table):
    """ Prune any files which are missing from the specified table """
    LOGGER.debug("Pruning missing %s files", table.__name__)
    removed_paths: typing.List[str] = []

    @orm.db_session(retry=5)
    def fill():
        try:
            for item in table.select():
                if not os.path.isfile(item.file_path):
                    LOGGER.info("%s disappeared: %s", table.__name__, item.file_path)
                    removed_paths.append(item.file_path)
        except Exception:  # pylint:disable=broad-except
            LOGGER.exception("Error pruning %s", table.__name__)

    @orm.db_session(retry=5)
    def kill(path):
        LOGGER.debug("Pruning %s %s", table.__name__, path)
        try:
            item = table.get(file_path=path)
            if item and not os.path.isfile(item.file_path):
                item.delete()
        except Exception:  # pylint:disable=broad-except
            LOGGER.exception("Error pruning %s", table.__name__)

    fill()
    for item in removed_paths:
        kill(item)


def scan_index(content_dir, wait_start=True):
    """ Scan all files in a content directory """
    LOGGER.debug("Reindexing content from %s", content_dir)

    from .flask_wrapper import current_app
    indexer = current_app.indexer

    def scan_directory(root, files):
        """ Helper function to scan a single directory """
        LOGGER.debug("scanning directory %s", root)
        try:
            for file in files:
                fullpath = os.path.join(root, file)
                relpath = os.path.relpath(fullpath, content_dir)

                if not is_scannable(fullpath):
                    continue

                fingerprint = utils.file_fingerprint(fullpath)
                last_fingerprint = get_last_fingerprint(fullpath)
                if fingerprint != last_fingerprint:
                    LOGGER.debug("%s: %s -> %s", fullpath, last_fingerprint, fingerprint)
                    indexer.scan_file(fullpath, relpath, 0, wait_start)
        except Exception:  # pylint:disable=broad-except
            LOGGER.exception("Got error parsing directory %s", root)

    for root, _, files in os.walk(content_dir, followlinks=True):
        indexer.submit(scan_directory, root, files)

    for table in (model.Entry, model.Category, model.Image, model.FileFingerprint):
        indexer.submit(prune_missing, table)



""" Periodic maintenance tasks """

import time
import typing


class Maintenance:
    """ Container for periodic maintenance tasks """

    def __init__(self, app):
        self.app = app
        self.tasks: typing.Dict[typing.Callable[[], None],
                                typing.Dict[str, float]] = {}

    def register(self, func: typing.Callable[[], None], interval: float):
        """ Registers a task to run periodically """
        self.tasks[func] = {'interval': interval}

    def run(self, force: bool = False):
        """ Run all pending tasks; 'force' will run all tasks whether they're
        pending or not. """
        with self.app.app_context():
            now = time.time()
            for func, spec in self.tasks.items():
                if force or now >= spec.get('next_run', 0):
                    func()
                    spec['next_run'] = now + spec['interval']



""" the Flask wrapper class """

import functools
import logging
import re
import typing

import arrow
import authl.flask
import flask
import werkzeug.exceptions
from werkzeug.utils import cached_property

# pylint:disable=cyclic-import
from . import (caching, cli, config, html_entry, image, index, maintenance,
               model, rendering, search, tokens, user, utils, view)

LOGGER = logging.getLogger(__name__)


class Publ(flask.Flask):
    """ A Publ application.

    At present, only one Publ application can be instanced at one time.
    """

    _instance = None

    def __init__(self, name, cfg, **kwargs):
        """ Constructor for a Publ application. Accepts the following parameters:

        :param str name: The name of the app
        :param str cfg: Application configuration

        Additional keyword arguments are forwarded along to the Flask constructor.

        Configuration keys:

        * ``database_config``: The database confugiration to be provided to PonyORM.
            See https://docs.ponyorm.org/database.html for more information
        * ``content_folder``: The folder that stores the site content
        * ``template_folder``: The folder that contains the Jinja templates
        * ``static_folder``: The folder that contains static content
        * ``static_url_path``: The URL mount point for the static content folder
        * ``image_output_subdir``: The subdirectory of the static content folder to
            store the image rendition cache
        * ``index_rescan_interval``: How frequently (in seconds) to rescan the
            content index
        * ``index_wait_time``: How long to wait (in seconds) before starting to
            process content updates
        * ``image_cache_interval``: How frequently (in seconds) to clean up the
            image rendition cache
        * ``image_cache_age``: The maximum age (in seconds) of an image rendition
        * ``timezone``: The site's local time zone
        * ``cache``: Page render cache configuration; see
            https://flask-caching.readthedocs.io/en/latest/#configuring-flask-caching
        * ``markdown_extensions``: The extensions to enable by default for the
            Markdown processing library. See https://misaka.61924.nl/#extensions
            for details
        * ``auth``: Authentication configuration. See the `Authl configuration
            documentation <https://authl.readthedocs.io>`_. Additionally, setting
            the key ``AUTH_FORCE_HTTPS`` to a truthy value can be used to force the
            user to switch to a secure connection when they log in, and setting
            ``AUTH_TOKEN_STORAGE`` to a :py:class:`authl.tokens.TokenStore` will
            use that for token storage instead of the default.
        * ``user_list``: The file that configures the user and group list
        * ``admin_group``: The user or group that has full administrative access
            to all entries regardless of permissions
        * ``auth_log_prune_interval``: How frequently to prune the authentication log, in seconds
        * ``auth_log_prune_age``: How long to retain authentication log entries, in seconds
        * ``ticket_lifetime``: How long an IndieAuth ticket lasts, in seconds
        * ``token_lifetime``: How long an IndieAuth token lasts, in seconds
        """
        # pylint:disable=too-many-branches,too-many-statements

        if Publ._instance and Publ._instance is not self:
            LOGGER.warning("Only one Publ app can run at a time (%s,%s)", Publ._instance, self)
        Publ._instance = self

        super().__init__(name,
                         template_folder=cfg.get('template_folder', 'templates'),
                         static_folder=cfg.get('static_folder', 'static'),
                         static_url_path=cfg.get('static_url_path', '/static'),
                         **kwargs)

        self.publ_config = config.Config(cfg)

        if 'AUTH_FORCE_SSL' in self.publ_config.auth:
            raise ValueError('AUTH_FORCE_SSL is deprecated; use AUTH_FORCE_HTTPS instead')

        self._regex_map = []

        self.url_map.converters['category'] = utils.CategoryConverter
        self.url_map.converters['template'] = utils.TemplateConverter

        for route in [
                '/',
                '/<category:category>/',
                '/<template:template>',
                '/<category:category>/<template:template>',
        ]:
            self.add_url_rule(route, 'category', rendering.render_category)

        for route in [
                '/<int:entry_id>',
                '/<int:entry_id>-',
                '/<int:entry_id>-<slug_text>',
                '/<category:category>/<int:entry_id>',
                '/<category:category>/<int:entry_id>-',
                '/<category:category>/<int:entry_id>-<slug_text>',
        ]:
            self.add_url_rule(route, 'entry', rendering.render_entry)

        self.add_url_rule('/<path:path>.PUBL_PATHALIAS',
                          'path_alias', rendering.render_path_alias)

        self.add_url_rule('/_async/<path:filename>',
                          'async', image.get_async)

        self.add_url_rule('/_', 'chit', rendering.render_transparent_chit)

        self.add_url_rule('/_file/<path:filename>',
                          'asset', rendering.retrieve_asset)

        self.config['TRAP_HTTP_EXCEPTIONS'] = True
        self.register_error_handler(
            werkzeug.exceptions.HTTPException, rendering.render_exception)

        self.jinja_env.globals.update(  # pylint: disable=no-member
            get_view=view.get_view,
            arrow=arrow,
            static=utils.static_url,
            get_template=rendering.get_template,
            login=utils.auth_link('login'),
            logout=utils.auth_link('logout'),
            token_endpoint=utils.CallableProxy(lambda: utils.secure_link('tokens')),
            secure_url=utils.secure_link,
        )

        self.jinja_env.filters['strip_html'] = html_entry.strip_html  # pylint: disable=no-member

        caching.init_app(self, self.publ_config.cache)

        def logout(redir=''):
            """ Log out from the thing """
            if flask.request.method == 'POST':
                LOGGER.info("Logging out %s", flask.session.get('me'))
                LOGGER.info("Redir: %s", redir)
                LOGGER.info("Request path: %s", flask.request.path)

                flask.session.pop('me')
                return flask.redirect('/' + redir)

            tmpl = rendering.map_template('/', 'logout')
            return rendering.render_publ_template(tmpl)[0]

        if self.publ_config.auth:
            for route in [
                    '/_logout',
                    '/_logout/',
                    '/_logout/<path:redir>'
            ]:
                self.add_url_rule(route, 'logout', logout, methods=['GET', 'POST'])

            for route in [
                    '/_admin',
                    '/_admin/<by>'
            ]:
                self.add_url_rule(route, 'admin', rendering.admin_dashboard)

            self.add_url_rule('/_tokens', 'tokens', tokens.indieauth_endpoint,
                              methods=['GET', 'POST'])

            self.before_request(user.log_user)

            def add_token_endpoint(response):
                endpoint = utils.secure_link("tokens", _external=True)
                response.headers.add(
                    'Link',
                    f'<{endpoint}>; rel="token_endpoint"')
                return response

            self.after_request(add_token_endpoint)

            # Force the authl instance to load before the first request, after the
            # app has had a chance to set secret_key
            self.before_first_request(lambda: self.authl)
        else:
            # Auth isn't configured, so make some placeholder routes so that
            # url_for doesn't fail and reasonable errors get raised
            def no_auth(redir=''):
                raise werkzeug.exceptions.NotFound()

            for base in ('_login', '_logout'):
                for route in (
                    f'/{base}',
                    f'/{base}/',
                    f'/{base}/<path:redir>',
                ):
                    self.add_url_rule(route, 'login', no_auth, methods=['GET', 'POST'])

        self._maint = maintenance.Maintenance(self)
        self.indexer = index.Indexer(self, self.publ_config.index_wait_time)

        if self.publ_config.index_rescan_interval:
            self._maint.register(functools.partial(index.scan_index,
                                                   self.publ_config.content_folder),
                                 self.publ_config.index_rescan_interval)

        if self.publ_config.image_cache_interval and self.publ_config.image_cache_age:
            self._maint.register(functools.partial(image.clean_cache,
                                                   self.publ_config.image_cache_age),
                                 self.publ_config.image_cache_interval)

        if self.publ_config.auth_log_prune_interval and self.publ_config.auth_log_prune_age:
            self._maint.register(functools.partial(user.prune_log,
                                                   self.publ_config.auth_log_prune_age),
                                 self.publ_config.auth_log_prune_interval)

        self.before_request(self._maint.run)

        if self.debug:
            # We're in debug mode so we don't want to scan until everything's up
            # and running
            self.before_first_request(self._startup)
        else:
            # In production, register the exception handler and scan the index
            # immediately
            self.register_error_handler(Exception, rendering.render_exception)
            self._startup()

        cli.setup(self)

    @cached_property
    def authl(self):
        """ Get the authl instance """
        if self.publ_config.auth:
            auth_force_https = self.publ_config.auth.get(
                'AUTH_FORCE_HTTPS',
                self.publ_config.auth.get('AUTH_FORCE_SSL'))
            if auth_force_https:
                self.config['SESSION_COOKIE_SECURE'] = True

            return authl.flask.AuthlFlask(
                self, self.publ_config.auth,
                login_path='/_login',
                login_name='login',
                callback_path='/_cb',
                tester_path='/_ct',
                force_https=bool(self.publ_config.auth.get('AUTH_FORCE_HTTPS')),
                login_render_func=rendering.render_login_form,
                on_verified=user.register,
                token_storage=self.publ_config.auth.get('AUTH_TOKEN_STORAGE'))
        return None

    def path_alias_regex(self, regex):
        r""" A decorator that adds a path-alias regular expression; calls
        add_path_regex.

        Example usage:

        @app.path_alias_regex(r'/d/([0-9]{8}(_w)?)\.php')
        def redirect_date(match):
            return flask.url_for('category', category='comics',
                                 date=match.group(1)), True
        """
        def decorator(func):
            """ Adds the function to the regular expression alias list """
            return self.add_path_regex(regex, func)
        return decorator

    def add_path_regex(self, regex, func):
        """ Add a path-alias regex callback to the request router. Takes the
        following arguments:

        regex -- The regular expression for the path-alias hook
        f -- A function taking a `re.match` object on successful match, and
            returns a tuple of `(url, is_permanent)`; url can be `None` if the
            function decides it should not redirect after all.

        The function may also use `flask.request.args` or the like if it needs
        to make a determination based on query args.
        """
        self._regex_map.append((regex, func))
        return func

    def test_path_regex(self, path):
        """ Evaluate the registered path-alias regular expressions. Returns the
        result of the first handler that successfully matches the path.
        """
        for regex, func in self._regex_map:
            match = re.match(regex, path)
            if match:
                dest, permanent = func(match)
                if dest:
                    return dest, permanent

        return None, None

    def _startup(self):
        """ Startup routine for initiating the content indexer """

        with self.app_context():
            model.setup(self.publ_config.database_config)

            self.search_index = search.SearchIndex(self.publ_config)
            self.jinja_env.globals.update(  # pylint: disable=no-member
                search=self.search_index.query,
            )

            import click

            ctx = click.get_current_context(silent=True)
            if not ctx or ctx.info_name == 'run':
                index.scan_index(self.publ_config.content_folder)
                if self.publ_config.index_enable_watchdog:
                    index.background_scan(self.publ_config.content_folder)


current_app = typing.cast(Publ, flask.current_app)  # pylint:disable=invalid-name



# links.py
""" Functions for manipulating outgoing HTML links """

import os
import re
import typing
from urllib.parse import urljoin

from flask import request

from . import entry  # pylint:disable=cyclic-import
from . import image, model, utils
from .config import config


def resolve(path: str, search_path: typing.Tuple[str, ...], absolute: bool = False) -> str:
    """ Remap a link or source target to an appropriate entry or image rendition """

    # Resolve external URLs
    if re.match(r'([a-z][a-z0-9+.\-]*:)?//', path, re.I):
        return path

    # Resolve static assets
    if path.startswith('@'):
        return utils.static_url(path[1:], absolute)

    path, sep, anchor = path.partition('#')

    # Resolve entries
    found = find_entry(path, search_path)
    if found:
        return entry.Entry.load(found).permalink(absolute=absolute) + sep + anchor

    # Resolve images and assets
    img_path, img_args, _ = image.parse_image_spec(path)
    img = image.get_image(img_path, search_path)
    if not isinstance(img, image.ImageNotFound):
        path, _ = img.get_rendition(**{**img_args, 'absolute': absolute})

    # We don't know what this is, so just treat it like a normal URL.
    if absolute:
        path = urljoin(request.url, path)

    return path + sep + anchor


def find_entry(rel_path: str, search_path: typing.Tuple[str, ...]) -> typing.Optional[model.Entry]:
    """ Find an entry by relative path. Arguments:

    rel_path -- the entry's filename (or entry ID)
    search_path -- a list of directories to check in

    Returns: the resolved Entry object
    """

    try:
        entry_id = int(rel_path)
        record = model.Entry.get(id=entry_id)
        if record:
            return record
    except ValueError:
        pass

    if rel_path.startswith('/'):
        search_path = (config.content_folder,)
        rel_path = '.' + rel_path

    for where in search_path:
        abspath = os.path.normpath(os.path.join(where, rel_path))
        record = model.Entry.get(file_path=abspath)
        if record:
            return record
    return None



# queries.py
""" Collection of commonly-used queries """

from enum import Enum

import arrow
from pony import orm

from . import model, utils


class InvalidQueryError(RuntimeError):
    """ An exception to raise if a query is invalid """


class FilterCombiner(Enum):
    """ Which operation to use when combining multiple operations in a filter """
    ANY = 0     # any of the criteria are present
    OR = 0      # synonym for ANY

    ALL = 1     # all of the criteria are present
    AND = 1     # synonym for ALL

    NONE = 2    # none of the criteria are present
    NOT = 2     # synonym for NONE


# Ordering queries for different sort orders
ORDER_BY = {
    'newest': (orm.desc(model.Entry.local_date), orm.desc(model.Entry.id)),
    'oldest': (model.Entry.local_date, model.Entry.id),
    'title': (model.Entry.sort_title, model.Entry.id)
}

REVERSE_ORDER_BY = {
    'newest': (model.Entry.local_date, model.Entry.id),
    'oldest': (orm.desc(model.Entry.local_date), orm.desc(model.Entry.id)),
    'title': (orm.desc(model.Entry.sort_title), orm.desc(model.Entry.id))
}


def where_entry_visible(query, timestamp=None):
    """ Generate a where clause for currently-visible entries

    Arguments:

    date -- The date to generate it relative to (defaults to right now)
    """

    ref_time = arrow.utcnow().float_timestamp if timestamp is None else timestamp

    return query.filter(lambda e:
                        e.status == model.PublishStatus.PUBLISHED.value or (
                            e.status == model.PublishStatus.SCHEDULED.value and (
                                e.utc_timestamp <= ref_time)
                        ))


def where_entry_visible_future(query):
    """ Generate a where clause for entries that are visible now or in the future """

    return query.filter(lambda e:
                        e.status in (model.PublishStatus.PUBLISHED.value,
                                     model.PublishStatus.SCHEDULED.value))


def where_entry_deleted(query):
    """ Generate a where clause for entries that have been deleted """
    return query.filter(lambda e:
                        e.status == model.PublishStatus.GONE.value)


def where_entry_category(query, category, recurse=False):
    """ Generate a where clause for a particular category """

    if utils.is_list(category):
        clist = {str(c) for c in category}

        if recurse:
            prefixes = tuple(c + '/' for c in clist)
            categories = orm.select(e.category for e in model.Entry)
            allowed = {*clist, *[c for c in categories if c.startswith(prefixes)]}
            return query.filter(lambda e: e.category in allowed)

        return query.filter(lambda e: e.category in clist)

    category = str(category)
    if category and recurse:
        # We're recursing and aren't in /, so add the prefix clause
        return query.filter(lambda e:
                            e.category == category or
                            e.category.startswith(category + '/'))

    if not recurse:
        # We're not recursing, so we need an exact match on a possibly-empty
        # category
        return query.filter(lambda e: e.category == category)

    # We're recursing and have no category, which means we're doing nothing
    return query


def where_entry_category_not(query, category):
    """ Generate a where clause for not being in a particular category """
    if utils.is_list(category):
        clist = [str(c) for c in category]
        return query.filter(lambda e: e.category not in clist)
    return query.filter(lambda e: e.category != str(category))


def where_before_entry(query, ref):
    """ Generate a where clause for prior entries

    ref -- The entry of reference
    """
    if not ref:
        raise InvalidQueryError("Attempted to reference non-existent entry")
    return query.filter(lambda e: (e.local_date, e.id) < (ref.local_date, ref.id))


def where_after_entry(query, ref):
    """ Generate a where clause for later entries

    ref -- the entry of reference
    """
    if not ref:
        raise InvalidQueryError("Attempted to reference non-existent entry")
    return query.filter(lambda e: (e.local_date, e.id) > (ref.local_date, ref.id))


def where_entry_last(query, ref):
    """ Generate a where clause where this is the last entry

    ref -- the entry of reference
    """
    if not ref:
        raise InvalidQueryError("Attempted to reference non-existent entry")
    # return query.filter(lambda e: (e.local_date, e.id) <= (ref.local_date, ref.id))
    # workaround for https://github.com/ponyorm/pony/issues/579
    return query.filter(lambda e: e.local_date < ref.local_date or
                        (e.local_date == ref.local_date and e.id <= ref.id))


def where_entry_first(query, ref):
    """ Generate a where clause where this is the first entry

    ref -- the entry of reference
    """
    if not ref:
        raise InvalidQueryError("Attempted to reference non-existent entry")
    # return query.filter(lambda e: (e.local_date, e.id) >= (ref.local_date, ref.id))
    # workaround for https://github.com/ponyorm/pony/issues/579
    return query.filter(lambda e: e.local_date > ref.local_date or
                        (e.local_date == ref.local_date and e.id >= ref.id))


def where_entry_type(query, entry_type):
    """ Generate a where clause for entries of certain types

    entry_type -- one or more entries to check against
    """
    if utils.is_list(entry_type):
        return query.filter(lambda e: e.entry_type in entry_type)
    return query.filter(lambda e: e.entry_type == entry_type)


def where_entry_type_not(query, entry_type):
    """ Generate a where clause for entries that aren't of certain types

    entry_type -- one or more entries to check against
    """
    if utils.is_list(entry_type):
        return query.filter(lambda e: e.entry_type not in entry_type)
    return query.filter(lambda e: e.entry_type != entry_type)


def where_entry_tag(query, tags, operation: FilterCombiner):
    """ Generate a where clause for entries with the given tags """
    tags = [utils.tag_key(t) for t in utils.as_list(tags)]

    if operation == FilterCombiner.ANY:
        return query.filter(lambda e: orm.exists(t for t in e.tags
                                                 if t.tag.key in tags))

    if operation == FilterCombiner.ALL:
        for tag in tags:
            # pylint:disable=undefined-loop-variable,cell-var-from-loop
            query = query.filter(lambda e: orm.exists(t for t in e.tags
                                                      if t.tag.key == tag))
        return query

    if operation == FilterCombiner.NONE:
        for tag in tags:
            # pylint:disable=cell-var-from-loop
            query = query.filter(lambda e: not orm.exists(t for t in e.tags
                                                          if t.tag.key == tag))
        return query

    raise InvalidQueryError("Unsupported FilterCombiner " + str(operation))


def where_entry_date(query, datespec):
    """ Where clause for entries which match a textual date spec

    datespec -- The date spec to check for, in YYYY[[-]MM[[-]DD]] format
    """
    try:
        date, interval, _ = utils.parse_date(datespec)
    except ValueError as error:
        raise InvalidQueryError(f"Invalid date {datespec}") from error
    start_date, end_date = date.span(interval)

    return query.filter(lambda e:
                        e.local_date >= start_date.naive and
                        e.local_date <= end_date.naive
                        )


def where_entry_attachments(query, entry):
    """ Where clause for entries which are attachments of the specified one. """
    return query.filter(lambda e: e in entry.attachments)


def where_entry_attached(query, entry):
    """ Where clause for entries which this one is attached onto. """
    return query.filter(lambda e: e in entry.attached)


def where_entry_has_attachments(query, val):
    """ Where clause for entries which have attachments """
    return query.filter(lambda e: val == orm.exists(e.attachments))


def where_entry_is_attached(query, val):
    """ Where clause for entries which are attached """
    return query.filter(lambda e: val == orm.exists(e.attached))


def get_entry(entry):
    """ Helper function to get an entry by ID or by object """

    if isinstance(entry, model.Entry):
        return entry

    if hasattr(entry, '_record'):
        return getattr(entry, '_record')

    if isinstance(entry, (int, str)):
        try:
            return model.Entry.get(id=int(entry))
        except ValueError as error:
            raise InvalidQueryError(f"Invalid entry ID {entry}") from error

    raise InvalidQueryError(f"entry is of unknown type {type(entry)}")


def build_query(spec):
    """ build the where clause based on a view specification

    spec -- The view specification. Contains the following possible values:
        future -- Boolean; whether to include entries from the future
        category -- Which category to limit to
        recurse -- Whether to include subcategories
        entry_type -- one or more entry types to include
        entry_type_not -- one or more entry types to exclude
        date -- a date spec
        last -- the last entry to end a view on
        first -- the first entry to start a view on
        before -- get entries from before this one
        after -- get entries from after this one
    """
    # pylint:disable=too-many-branches

    query = model.Entry.select()

    # primarily restrict by publication status
    if spec.get('_all', False):
        pass
    elif spec.get('_deleted', False):
        query = where_entry_deleted(query)
    else:
        if spec.get('future', False):
            query = where_entry_visible_future(query)
        else:
            query = where_entry_visible(query)

    # restrict by category
    if spec.get('category') is not None:
        recurse = spec.get('recurse', False)
        query = where_entry_category(query, spec.get('category', ''), recurse)

    if spec.get('category_not') is not None:
        query = where_entry_category_not(query, spec.get('category_not'))

    if spec.get('entry_type') is not None:
        query = where_entry_type(query, spec['entry_type'])

    if spec.get('entry_type_not') is not None:
        query = where_entry_type_not(query, spec['entry_type_not'])

    if spec.get('tag') is not None:
        query = where_entry_tag(query, spec['tag'],
                                FilterCombiner[spec.get('tag_filter', 'ANY').upper()])

    if spec.get('date') is not None:
        query = where_entry_date(query, spec['date'])

    if spec.get('last') is not None:
        query = where_entry_last(query, get_entry(spec['last']))

    if spec.get('first') is not None:
        query = where_entry_first(query, get_entry(spec['first']))

    if spec.get('before') is not None:
        query = where_before_entry(query, get_entry(spec['before']))

    if spec.get('after') is not None:
        query = where_after_entry(query, get_entry(spec['after']))

    if spec.get('attachments') is not None:
        query = where_entry_attachments(query, get_entry(spec['attachments']))

    if spec.get('attached') is not None:
        query = where_entry_attached(query, get_entry(spec['attached']))

    if spec.get('has_attachments') is not None:
        query = where_entry_has_attachments(query, spec['has_attachments'])

    if spec.get('is_attached') is not None:
        query = where_entry_is_attached(query, spec['is_attached'])

    return query.distinct()




""" Authenticated user functionality """

import ast
import collections
import configparser
import datetime
import logging
import typing
import urllib.parse

import arrow
import authl.disposition
import flask
import werkzeug.exceptions as http_error
from pony import orm
from werkzeug.utils import cached_property

from . import caching, model, tokens, utils
from .config import config

LOGGER = logging.getLogger(__name__)


@caching.cache.memoize(timeout=5)
def get_groups(identity: str, include_self: bool = True) -> typing.Set[str]:
    """ Get the group membership for the given identity """

    @caching.cache.memoize(timeout=10)
    def load_groups() -> typing.DefaultDict[str, typing.Set[str]]:
        # We only want empty keys; \000 is unlikely to turn up in a well-formed text file
        cfg = configparser.ConfigParser(delimiters=(
            '\000'), allow_no_value=True, interpolation=None)
        # disable authentication lowercasing; identities should only be case-desensitized
        # by the auth backend
        cfg.optionxform = lambda option: option  # type:ignore
        cfg.read(config.user_list)

        groups: typing.DefaultDict[str, typing.Set[str]] = collections.defaultdict(set)

        # populate the group list for each member
        for group, members in cfg.items():
            for member in members.keys():
                groups[member].add(group)

        return groups

    groups = load_groups()
    result: typing.Set[str] = set()
    pending: typing.Deque[str] = collections.deque()

    pending.append(identity)

    while pending:
        check = pending.popleft()
        if check not in result:
            if include_self or check != identity:
                result.add(check)
            pending += groups.get(check, [])

    return result


class User(caching.Memoizable):
    """ An authenticated user """

    def __init__(self, identity: str,
                 auth_type: typing.Optional[str] = None,
                 scope: typing.Optional[str] = None):
        """ Initialize the user object.

        :param str identity: The federated identity name
        :param str auth_type: The authentication mechanism
        :param str scope: The user's access scope
        """

        self._identity = identity
        self._auth_type = auth_type
        self._scope = scope

    def _key(self):
        return self._identity, self._auth_type, self._scope

    def __lt__(self, other):
        return self.identity < other.identity

    @cached_property
    def identity(self):
        """ The federated identity name of the user """
        return self._identity

    @cached_property
    def humanize(self) -> str:
        """ A humanized version of the identity string """
        url = self.profile.get('profile_url', self._identity)
        parsed = urllib.parse.urlparse(url)
        return ''.join(p for p in (
            f'{parsed.scheme}:' if parsed.scheme not in ('http', 'https') else '',
            parsed.netloc,
            parsed.path,
        ))

    @cached_property
    def name(self) -> str:
        """ The readable name of the user """
        if 'name' in self.profile:
            return self.profile['name']

        return self.humanize

    @property
    def profile(self) -> dict:
        """ Get the user's profile """
        return self._info[0]

    @cached_property
    def groups(self) -> typing.Set[str]:
        """ The group memberships of the user, for display purposes """
        return get_groups(self._identity, False)

    @cached_property
    def auth_groups(self) -> typing.Set[str]:
        """ The group memberships of the user, for auth purposes """
        return get_groups(self._identity, True)

    @cached_property
    def is_admin(self) -> bool:
        """ Returns whether this user has administrator permissions """
        return bool(config.admin_group and config.admin_group in self.auth_groups)

    @cached_property
    def auth_type(self):
        """ The type of user authentication (session, token, etc.) """
        return self._auth_type

    @cached_property
    def scope(self):
        """ The permission scope of the user """
        return self._scope

    @cached_property
    def _info(self) -> typing.Tuple[dict,
                                    typing.Optional[datetime.datetime],
                                    typing.Optional[datetime.datetime]]:
        """ Gets the user info from the database

        :returns: profile, last_login, last_seen
        """
        with orm.db_session():
            record = model.KnownUser.get(user=self._identity)
            if record:
                return (record.profile.copy(),
                        record.last_login,
                        record.last_seen)
        return {}, None, None

    @property
    def last_login(self) -> typing.Optional[arrow.Arrow]:
        """ Get the latest known login time for the user, if any """
        date = self._info[1]
        return arrow.get(date).to(config.timezone) if date else None

    @property
    def last_seen(self) -> arrow.Arrow:
        """ Get the latest known active time for the user """
        date = self._info[2]
        return arrow.get(date).to(config.timezone) if date else None

    def token(self, lifetime: int, scope: str = None) -> str:
        """ Get a bearer token for this user """
        return tokens.get_token(self.identity, lifetime, scope)


@utils.stash
def get_active() -> typing.Optional[User]:
    """ Get the active user """
    if 'token_error' in flask.g:
        # We already got an error trying to parse an access token
        return None

    if 'Authorization' in flask.request.headers:
        token = tokens.parse_authorization_header(flask.request.headers['Authorization'])
        try:
            return User(token['me'], 'token', token.get('scope'))
        except http_error.HTTPException as error:
            flask.g['token_error'] = error.description
            raise

    if flask.session.get('me'):
        return User(flask.session['me'], 'session')

    return None


@orm.db_session(retry=5)
def log_access(record, cur_user, authorized):
    """ Log a user's access to the audit log """
    LOGGER.info("log_access %s %s %s", record, cur_user, authorized)
    if cur_user:
        values = {
            'date': arrow.utcnow().datetime,
            'entry': record,
            'authorized': authorized,
            'user': cur_user.identity,
            'user_groups': str(cur_user.groups) if cur_user.groups else ''
        }
        log_entry = model.AuthLog.get(entry=record, user=cur_user.identity)
        if log_entry:
            log_entry.set(**values)
        else:
            model.AuthLog(**values)


def _get_group_set(groups):
    try:
        return ast.literal_eval(groups)
    except (ValueError, SyntaxError):
        return set(groups.split(',')) if groups else set()


@orm.db_session(retry=5)
def register(verified: authl.disposition):
    """ Registers a user from the on_verified Authl hook """
    if isinstance(verified, authl.disposition.Verified):
        LOGGER.info("Got login from user %s with profile %s", verified.identity, verified.profile)
        identity = utils.canonicize_url(verified.identity)
        now = arrow.utcnow().datetime
        values = {
            'last_login': now,
            'last_seen': now,
            'profile': verified.profile
        }

        record = model.KnownUser.get(user=identity)
        if record:
            record.set(**values)
        else:
            record = model.KnownUser(user=identity, **values)

        if (verified.profile
            and 'endpoints' in verified.profile
                and 'ticket_endpoint' in verified.profile['endpoints']):
            tokens.send_auth_ticket(identity, flask.request.url_root,
                                    verified.profile['endpoints']['ticket_endpoint'])


@orm.db_session(retry=5)
def log_user():
    """ Update the user table to see who's been by """
    identity = flask.session.get('me')
    if identity:
        values = {
            'last_seen': arrow.utcnow().datetime,
        }

        record = model.KnownUser.get(user=identity)
        if record:
            record.set(**values)
        else:
            record = model.KnownUser(user=identity, **values)


@orm.db_session()
def known_users(days=None):
    """ Get the users known to the system, as a list of (user,last_seen) """
    query = model.KnownUser.select()
    if days:
        since = (arrow.utcnow() - datetime.timedelta(days=days)).datetime
        query = orm.select(e for e in query if e.last_seen >= since)

    query = query.order_by(orm.desc(model.KnownUser.last_seen))

    return [(User(record.user), arrow.get(record.last_seen).to(config.timezone))
            for record in query]


LogEntry = collections.namedtuple(
    'LogEntry', ['date', 'entry', 'user', 'user_groups', 'authorized'])


@orm.db_session()
def auth_log(start=0, count=100):
    """ Get the logged accesses to each entry """
    query = model.AuthLog.select().order_by(orm.desc(model.AuthLog.date))[start:]

    return [LogEntry(date=arrow.get(record.date).to(config.timezone),
                     entry=record.entry,
                     user=User(record.user),
                     user_groups=_get_group_set(record.user_groups),
                     authorized=record.authorized)
            for record in query[:count]], len(query) - count


@orm.db_session(optimistic=False)
def prune_log(seconds):
    """ Delete log entries which are older than the cutoff """
    since = (arrow.utcnow() - datetime.timedelta(seconds=seconds)).datetime
    LOGGER.debug("Purging auth entries older than %s", since)
    orm.delete(e for e in model.AuthLog if e.date < since)
    orm.delete(e for e in model.KnownUser if e.last_seen < since)



# html_entry.py
""" HTML entry processing functionality """

import html
import logging
import re
import typing

import flask

from . import image, links, utils
from .config import config

LOGGER = logging.getLogger(__name__)


class HTMLEntry(utils.HTMLTransform):
    """ An HTML manipulator to fixup src and href attributes """

    def __init__(self, args, search_path):
        super().__init__()

        self._search_path = search_path
        self._config = args

    def handle_decl(self, decl):
        self.append('<!' + decl + '>')

    def handle_data(self, data):
        self.append(data)

    def handle_starttag(self, tag, attrs):
        """ Handle a start tag """
        self._handle_tag(tag, attrs, False)

    def handle_endtag(self, tag):
        """ Handle an end tag """
        self.append('</' + tag + '>')

    def handle_pi(self, data):
        LOGGER.warning("handle_pi: '%s'", data)

    def handle_comment(self, data):
        self.append('<!--' + data + '-->')

    def handle_startendtag(self, tag, attrs):
        """ Handle a self-closing tag """
        self._handle_tag(tag, attrs, True)

    def _handle_tag(self, tag, attrs, self_closing):
        """ Handle a tag.

        attrs -- the attributes of the tag
        self_closing -- whether this is self-closing
        """

        if tag.lower() == 'img':
            attrs = self._image_attrs(attrs)

        # Remap the attributes
        out_attrs = []
        for key, val in attrs:
            if (key.lower() == 'href'
                    or key.startswith('$')
                    or (key.lower() == 'src' and tag.lower() != 'img')):
                if key.startswith('$'):
                    key = key[1:]
                out_attrs.append((key, links.resolve(
                    val, self._search_path, self._config.get('absolute'))))
            else:
                out_attrs.append((key, val))

        self.append(
            utils.make_tag(
                tag,
                out_attrs,
                self_closing))

    def _image_attrs(self, attrs):
        """ Rewrite the SRC attribute on an <img> tag, possibly adding a SRCSET.
        """

        path = None
        attr_args = {}
        is_rewritten = False

        for key, val in attrs:
            if key.lower() == 'width' or key.lower() == 'height':
                try:
                    attr_args[key.lower()] = int(val)
                except ValueError:
                    pass
            elif key.lower() == 'src':
                path = val
            elif key == 'data-publ-rewritten':
                is_rewritten = True
                break

        if is_rewritten:
            # This img tag was already rewritten by a previous processor, so just
            # remove that attribute and return the tag's original attributes
            LOGGER.debug("Detected already-rewritten image; %s", attrs)
            return [(key, val) for key, val in attrs if key != 'data-publ-rewritten']

        if not path:
            # this img doesn't have a src attribute, so there's something unconventional going on
            return attrs

        img_path, img_args, _ = image.parse_image_spec(path)
        img = image.get_image(img_path, self._search_path)

        try:
            img_attrs = img.get_img_attrs({**self._config,
                                           **attr_args,
                                           **img_args})
        except FileNotFoundError as error:
            img_attrs = {
                'data-publ-error': f'File Not Found: {error.filename}'
            }
        except Exception as error:  # pylint:disable=broad-except
            LOGGER.exception("Got exception: %s", error)
            img_attrs = {
                'data-publ-error': f'Error: {error}'
            }

        # return the original attr list with the computed overrides in place
        return [(key, val) for key, val in attrs
                if key.lower() not in img_attrs] + list(img_attrs.items())


def process(text, args, search_path):
    """ Process an HTML entry's HTML """
    args = {**config.layout, **args}
    processor = HTMLEntry(args, search_path)
    processor.feed(text)
    text = processor.get_data()

    if not args.get('markup', True):
        text = strip_html(text)

    return flask.Markup(text)


class HTMLStripper(utils.HTMLTransform):
    """ Strip all HTML tags from a document, except those which are allowed """

    def __init__(self,
                 allowed_tags: typing.Iterable[str] = None,
                 allowed_attrs: typing.Iterable[str] = None,
                 remove_elements: typing.Iterable[str] = None):
        super().__init__()
        self._allowed_tags = set(utils.as_list(allowed_tags))
        self._allowed_attrs = set(utils.as_list(allowed_attrs))
        self._remove_elements = set(utils.as_list(remove_elements))
        self._remove_depth = 0

    def _filter(self, tag, attrs, **kwargs) -> str:
        if self._allowed_tags and tag in self._allowed_tags:
            return utils.make_tag(tag,
                                  {key: val
                                   for key, val in attrs
                                   if key in self._allowed_attrs},
                                  **kwargs)
        if self._remove_elements and tag in self._remove_elements:
            self._remove_depth += 1
        return ''

    def handle_starttag(self, tag, attrs):
        self.append(self._filter(tag, attrs))

    def handle_endtag(self, tag):
        if self._allowed_tags and tag in self._allowed_tags:
            self.append(f'</{tag}>')
        elif self._remove_elements and tag in self._remove_elements:
            self._remove_depth -= 1

    def handle_startendtag(self, tag, attrs):
        self.append(self._filter(tag, attrs, start_end=True))

    def handle_data(self, data):
        if not self._remove_depth:
            self.append(data)


def strip_html(text,
               allowed_tags: typing.Iterable[str] = None,
               allowed_attrs: typing.Iterable[str] = None,
               remove_elements: typing.Iterable[str] = None) -> str:
    """ Strip all HTML formatting off of a chunk of text """
    strip = HTMLStripper(allowed_tags, allowed_attrs, remove_elements)
    strip.feed(str(text))
    return re.sub(r' +', r' ', html.unescape(strip.get_data())).strip()


class FirstParagraph(utils.HTMLTransform):
    """ Get just the first paragraph out of an HTML document """

    def __init__(self):
        super().__init__()

        self._consume = True
        self._found = False

        self._tag_stack = []

    def handle_starttag(self, tag, attrs):
        if tag.lower() == 'table':
            self._consume = False

        if tag.lower() == 'p':
            if self._found:
                self._consume = False

        self._tag_stack.append(tag)

        if self._consume:
            self.append(utils.make_tag(tag, attrs))

    def handle_endtag(self, tag):
        while self._tag_stack and self._tag_stack.pop() != tag:
            pass

        if self._consume:
            self.append(f'</{tag}>')

        if tag.lower() == 'table' and not self._found:
            self._consume = True

        if (not self._tag_stack or tag.lower() == 'p') and self._found:
            self._consume = False

    def handle_startendtag(self, tag, attrs):
        if self._consume:
            self.append(utils.make_tag(tag, attrs, True))

    def handle_data(self, data):
        if self._consume and data.strip():
            self._found = True
            self.append(data)


def first_paragraph(text):
    """ Extract the first paragraph of text from an HTML document """
    first_para = FirstParagraph()
    first_para.feed(str(text))
    text = first_para.get_data()
    return re.sub(r'<p> *</p>', r'', text)



""" Publ: A Flask-based site management system.

Like a static publishing system, but dynamic! See http://publ.beesbuzz.biz
for more information. """

import logging

from .flask_wrapper import Publ

LOGGER = logging.getLogger(__name__)


def publ(name, cfg):
    """ Legacy function that originally did a lot more """
    LOGGER.warning("This function is deprecated; use publ.Publ instead")
    return Publ(name, cfg)



""" Functions for manipulating images stored within local content """

import concurrent.futures
import functools
import logging
import os
import threading
import time
import typing

import flask
import PIL.Image
import slugify
from atomicwrites import atomic_write
from werkzeug.utils import cached_property

from .. import utils
from ..config import config
from .image import Image

LOGGER = logging.getLogger(__name__)

SizeType = typing.Tuple[int, int]
BoxType = typing.Tuple[int, int, int, int]
SizeSpecType = typing.Tuple[SizeType, typing.Optional[BoxType]]
PipelineEntry = typing.Tuple[typing.Optional[str],
                             typing.Optional[typing.Callable]]
ProcessingPipeline = typing.List[PipelineEntry]


def fix_orientation(image: PIL.Image) -> PIL.Image:
    """ adapted from https://stackoverflow.com/a/30462851/318857

        Apply Image.transpose to ensure 0th row of pixels is at the visual
        top of the image, and 0th column is the visual left-hand side.
        Return the original image if unable to determine the orientation.

        As per CIPA DC-008-2012, the orientation field contains an integer,
        1 through 8. Other values are reserved.
    """

    exif_orientation_tag = 0x0112
    exif_transpose_sequences = [
        [],
        [],
        [PIL.Image.FLIP_LEFT_RIGHT],
        [PIL.Image.ROTATE_180],
        [PIL.Image.FLIP_TOP_BOTTOM],
        [PIL.Image.FLIP_LEFT_RIGHT, PIL.Image.ROTATE_90],
        [PIL.Image.ROTATE_270],
        [PIL.Image.FLIP_TOP_BOTTOM, PIL.Image.ROTATE_90],
        [PIL.Image.ROTATE_90],
    ]

    try:
        # pylint:disable=protected-access
        orientation = image._getexif()[exif_orientation_tag]
        sequence = exif_transpose_sequences[orientation]
        return functools.reduce(type(image).transpose, sequence, image)
    except (TypeError, AttributeError, KeyError):
        # either no EXIF tags or no orientation tag
        pass
    return image


class LocalImage(Image):
    """ The basic Image class, which knows about the base version and how to
    generate renditions from it """

    _thread_pool = None

    @staticmethod
    def thread_pool():
        """ Get the rendition threadpool """
        # pylint:disable=consider-using-with
        if not LocalImage._thread_pool:
            LOGGER.info("Starting LocalImage threadpool")
            LocalImage._thread_pool = concurrent.futures.ThreadPoolExecutor(
                thread_name_prefix="Renderer",
                max_workers=config.image_render_threads)
        return LocalImage._thread_pool

    def __init__(self, record, search_path):
        """ Get the base image from an index record """
        super().__init__(search_path)

        self._record = record
        self._lock = threading.Lock()

    def _key(self):
        return LocalImage, self._record

    @property
    def _filename(self):
        return os.path.basename(self._record.file_path)

    def _get_rendition(self, output_scale=1, **kwargs):
        """ implements get_rendition and returns tuple of out_rel_path,size,pending """
        basename, ext = os.path.splitext(
            os.path.basename(self._record.file_path))
        basename = slugify.slugify(basename)

        if kwargs.get('format'):
            ext = '.' + kwargs['format']

        pipeline, out_args, size = self._build_pipeline(basename, ext, output_scale, kwargs)

        # Build the output filename
        out_basename = '_'.join([str(name) for name, _ in pipeline if name]) + ext
        out_rel_path = os.path.join(
            config.image_output_subdir,
            self._record.checksum[0:2],
            self._record.checksum[2:6],
            out_basename)
        out_fullpath = os.path.join(config.static_folder, out_rel_path)

        if os.path.isfile(out_fullpath):
            LOGGER.debug("rendition %s already exists", out_fullpath)
            os.utime(out_fullpath)
            pending = False
        else:
            LOGGER.debug("scheduling %s for render", out_fullpath)
            LocalImage.thread_pool().submit(
                self._render, out_fullpath, [op for _, op in pipeline if op], out_args)
            pending = True

        return out_rel_path, size, pending

    def _build_pipeline(self, basename, ext, output_scale, kwargs) -> typing.Tuple[
            ProcessingPipeline, typing.Dict, SizeType]:
        """
        Build the image processing pipeline

        Returns a tuple of ProcessingPipeline, output_args, size
        """

        pipeline: ProcessingPipeline = []

        # set the image basename
        pipeline.append((basename, None))
        pipeline.append((self._record.checksum[-10:], None))

        # Fix the EXIF orientation
        pipeline.append((None, fix_orientation))

        stash = {}
        out_args = {}
        if ext in ('.png', '.jpg', '.jpeg'):
            out_args['optimize'] = True

        # convert to RGBA
        def to_rgba(image):
            if image.mode == 'P':
                stash['paletted'] = True
                return image.convert('RGBA')
            return image

        pipeline.append((None, to_rgba))

        label: typing.Optional[str]

        # Set RGBA flattening options
        if (self._record.transparent and ext not in ('.png', '.gif')) or 'background' in kwargs:
            bg_color = kwargs.get('background')
            if isinstance(bg_color, (tuple, list)):
                label = 'b' + '-'.join([str(a) for a in bg_color])
            elif bg_color:
                label = 'b' + str(bg_color)
            else:
                label = None
            pipeline.append((label, lambda image: self.flatten(image, bg_color)))

        size, cropscale = self._build_pipeline_cropscale(output_scale, kwargs)
        if cropscale:
            pipeline.append(cropscale)

        # Set image quantization options
        if ext in ('.gif', '.png'):
            quantize = kwargs.get('quantize')

            def to_paletted(image):
                if ext == '.gif' or stash.get('paletted') or quantize:
                    return image.quantize(colors=quantize or 256)
                return image
            pipeline.append((f'q{quantize}' if quantize else None, to_paletted))

        # Set JPEG quality
        if ext in ('.jpg', '.jpeg') and kwargs.get('quality'):
            pipeline.append(('q' + str(kwargs['quality']), None))
            out_args['quality'] = kwargs['quality']

        return pipeline, out_args, size

    def _build_pipeline_cropscale(self, output_scale, kwargs) -> typing.Tuple[
            SizeType,
            typing.Optional[PipelineEntry]]:
        crop = utils.parse_tuple_string(kwargs.get('crop'))
        size, box = self.get_rendition_size(kwargs, output_scale, crop)

        if crop and box:
            # Both boxes are the same size; just line them up.
            box = (box[0] + crop[0], box[1] + crop[1],
                   box[2] + crop[0], box[3] + crop[1])
        elif crop:
            # We don't have a fit box, so just convert the crop box
            box = (crop[0], crop[1], crop[0] + crop[2], crop[1] + crop[3])

        # Apply the image cropscale
        if box or size[0] < self._record.width or size[1] < self._record.height:
            label = 'x'.join([str(v) for v in size])
            if box:
                label += '_' + '-'.join([str(v) for v in box])

            if 'scale_filter' in kwargs:
                try:
                    scale_filter = getattr(PIL.Image, kwargs['scale_filter'].upper())
                    label += f'f{scale_filter}'
                except AttributeError as error:
                    raise ValueError(
                        f"Invalid scale_filter value '{kwargs['scale_filter']}'") from error
            else:
                scale_filter = PIL.Image.LANCZOS
            return size, (label,
                          lambda image: image.resize(size=size,
                                                     box=box,
                                                     resample=scale_filter))
        return size, None

    def _render(self, path, operations: typing.List[typing.Callable], out_args):
        image = self._image

        with self._lock:
            if os.path.isfile(path):
                # file already exists
                return

            LOGGER.info("Rendering file %s", path)

            try:
                os.makedirs(os.path.dirname(path))
            except FileExistsError:
                pass

            try:
                for operation in operations:
                    image = operation(image)

                _, ext = os.path.splitext(path)
                with atomic_write(path, mode='w+b', suffix=ext, overwrite=True) as file:
                    image.save(file, **out_args)

                LOGGER.info("%s: complete", path)
            except Exception:  # pylint: disable=broad-except
                LOGGER.exception("Failed to render %s -> %s",
                                 self._record.file_path, path)

    def get_rendition(self, output_scale=1, **kwargs):
        """
        Get the rendition for this image, generating it if necessary.
        Returns a tuple of `(relative_path, width, height)`, where relative_path
        is relative to the static file directory (i.e. what one would pass into
        `get_static()`)

        output_scale -- the upsample factor for the requested rendition

        Keyword arguments:

        scale -- the downsample factor for the base rendition
        scale_min_width -- the minimum width after downsampling
        scale_min_height -- the minimum height after downsampling
        crop -- box to crop the original image into (left, top, right, bottom)
        width -- the width to target
        height -- the height to target
        max_width -- the maximum width
        max_height -- the maximum height
        resize -- how to fit the width and height; "fit", "fill", or "stretch"
        fill_crop_x -- horizontal offset fraction for resize="fill"
        fill_crop_y -- vertical offset fraction for resize="fill"
        format -- output format
        background -- background color when converting transparent to opaque
        quality -- the JPEG quality to save the image as
        quantize -- how large a palette to use for GIF or PNG images
        """
        out_rel_path, size, pending = self._get_rendition(output_scale, **kwargs)

        if pending:
            return flask.url_for('async',
                                 filename=out_rel_path,
                                 _external=kwargs.get('absolute')), size
        return utils.static_url(out_rel_path, kwargs.get('absolute')), size

    @cached_property
    def _image(self):
        with self._lock:
            image = PIL.Image.open(self._record.file_path)

        return image

    def get_rendition_size(self, spec, output_scale, crop) -> SizeSpecType:
        """
        Wrapper to determine the overall rendition size and cropping box

        Returns tuple of (size,box)
        """

        if crop:
            # Use the cropping rectangle size
            _, _, width, height = crop
        else:
            # Use the original image size
            width = self._record.width
            height = self._record.height

        mode = spec.get('resize', 'fit')

        if mode == 'fill':
            return self._get_rendition_fill_size(spec, width, height, output_scale)

        if mode == 'fit':
            return self._get_rendition_fit_size(spec, width, height, output_scale)

        if mode == 'stretch':
            return self._get_rendition_stretch_size(spec, width, height, output_scale)

        raise ValueError(f"Unknown resize mode {mode}")

    @staticmethod
    def _get_rendition_fit_size(spec, input_w, input_h, output_scale) -> SizeSpecType:
        """ Determine the scaled size based on the provided spec """

        width = input_w
        height = input_h

        scale = spec.get('scale')
        if scale:
            width = width / scale
            height = height / scale

        min_width = spec.get('scale_min_width')
        if min_width and width < min_width:
            height = height * min_width / width
            width = min_width

        min_height = spec.get('scale_min_height')
        if min_height and height < min_height:
            width = width * min_height / height
            height = min_height

        tgt_width, tgt_height = spec.get('width'), spec.get('height')

        if tgt_width and width > tgt_width:
            height = height * tgt_width / width
            width = tgt_width

        if tgt_height and height > tgt_height:
            width = width * tgt_height / height
            height = tgt_height

        tgt_width, tgt_height = spec.get('max_width'), spec.get('max_height')

        if tgt_width and width > tgt_width:
            height = height * tgt_width / width
            width = tgt_width

        if tgt_height and height > tgt_height:
            width = width * tgt_height / height
            height = tgt_height

        width = width * output_scale
        height = height * output_scale

        # Never scale to larger than the base rendition
        width = min(round(width), input_w)
        height = min(round(height), input_h)

        return (width, height), None

    @staticmethod
    def _get_rendition_fill_size(spec, input_w, input_h, output_scale) -> SizeSpecType:
        """ Determine the scale-crop size given the provided spec """

        width = input_w
        height = input_h

        scale = spec.get('scale')
        if scale:
            width = width / scale
            height = height / scale

        if spec.get('scale_min_width'):
            width = max(width, spec['spec_min_width'])

        if spec.get('scale_min_height'):
            height = max(height, spec['scale_min_height'])

        if spec.get('width'):
            width = min(width, spec['width'])
        if spec.get('max_width'):
            width = min(width, spec['max_width'])

        if spec.get('height'):
            height = min(height, spec['height'])
        if spec.get('max_height'):
            height = min(height, spec['max_height'])

        width = width * output_scale
        height = height * output_scale

        # Never scale to larger than the base rendition (but keep the output
        # aspect)
        if width > input_w:
            height = height * input_w / width
            width = input_w

        if height > input_h:
            width = width * input_h / height
            height = input_h

        # Determine the box size
        box_w = min(input_w, round(width * input_h / height))
        box_h = min(input_h, round(height * input_w / width))

        # Box offset
        box_x = round((input_w - box_w) * spec.get('fill_crop_x', 0.5))
        box_y = round((input_h - box_h) * spec.get('fill_crop_y', 0.5))

        return (round(width), round(height)), (box_x, box_y, box_x + box_w, box_y + box_h)

    @staticmethod
    def _get_rendition_stretch_size(spec, input_w, input_h, output_scale) -> SizeSpecType:
        """ Determine the scale-crop size given the provided spec """

        width = input_w
        height = input_h

        scale = spec.get('scale')
        if scale:
            width = width / scale
            height = height / scale

        min_width = spec.get('scale_min_width')
        if min_width and width < min_width:
            width = min_width

        min_height = spec.get('scale_min_height')
        if min_height and height < min_height:
            height = min_height

        tgt_width, tgt_height = spec.get('width'), spec.get('height')

        if tgt_width and width > tgt_width:
            width = tgt_width

        tgt_height = spec.get('height')
        if tgt_height and height > tgt_height:
            height = tgt_height

        tgt_width, tgt_height = spec.get('max_width'), spec.get('max_height')

        if tgt_width and width > tgt_width:
            width = tgt_width

        tgt_height = spec.get('height')
        if tgt_height and height > tgt_height:
            height = tgt_height

        width = width * output_scale
        height = height * output_scale

        return (round(width), round(height)), None

    @staticmethod
    def flatten(image, bgcolor=None):
        """ Flatten an image, with an optional background color """
        if bgcolor:
            background = PIL.Image.new('RGB', image.size, bgcolor)
            background.paste(image, mask=image.split()[3])
            image = background

        return image.convert('RGB')

    def _get_renditions(self, kwargs):
        """ Get a bunch of renditions; returns a tuple of 1x, 2x, size """
        img_1x, size = self.get_rendition(
            1, **utils.remap_args(kwargs, {"quality": "quality_ldpi"}))
        img_2x, _ = self.get_rendition(
            2, **utils.remap_args(kwargs, {"quality": "quality_hdpi"}))

        return (img_1x, img_2x, size)

    def _get_img_attrs(self, spec, style_parts):
        """ Get the attributes of an an <img> tag for this image, hidpi-aware """
        # pylint:disable=unused-argument

        # Get the 1x and 2x renditions
        img_1x, img_2x, size = self._get_renditions(spec)

        attrs = {
            'src': img_1x,
            'width': size[0],
            'height': size[1]
        }

        if img_1x != img_2x:
            attrs['srcset'] = f"{img_1x} 1x, {img_2x} 2x"

        return attrs

    def _css_background(self, **kwargs):
        """ Get the CSS specifiers for this as a hidpi-capable background image """

        # Get the 1x and 2x renditions
        img_1x, img_2x, _ = self._get_renditions(kwargs)

        tmpl = f'background-image: url("{img_1x}");'
        if img_2x and img_1x != img_2x:
            tmpl += f'background-image: image-set(url("{img_1x}") 1x, url("{img_2x}") 2x);'
        return tmpl

    @staticmethod
    def clean_cache(max_age):
        """ Clean the rendition cache of files older than max_age seconds """
        cache_dir = os.path.join(config.static_folder,
                                 config.image_output_subdir)
        return LocalImage.thread_pool().submit(LocalImage._clean_cache, max_age, cache_dir)

    @staticmethod
    def _clean_cache(max_age, cache_dir: str):
        threshold = time.time() - max_age
        LOGGER.debug("Deleting image renditions older than %d from %s", threshold,
                     cache_dir)

        # delete expired files
        for root, _, files in os.walk(cache_dir):
            for file in files:
                try:
                    path = os.path.join(root, file)
                    mtime = os.stat(path).st_mtime
                    LOGGER.debug("checking %s (%d)", path, mtime)
                    if os.path.isfile(path) and mtime < threshold:
                        os.unlink(path)
                        LOGGER.info("Expired stale rendition %s (mtime=%d threshold=%d)",
                                    path, mtime, threshold)
                except FileNotFoundError:
                    pass
            if next(os.scandir(root), None) is None:
                try:
                    os.removedirs(root)
                    LOGGER.info("Removed empty cache directory %s", root)
                except OSError:
                    LOGGER.exception("Couldn't remove %s", root)



# path_alias.py
""" Handling for URL aliases """

import logging
import typing
import urllib.parse

from flask import redirect, url_for
from pony import orm

from . import model, user

# redirection types
PERMANENT = 301
TEMPORARY = 302

LOGGER = logging.getLogger(__name__)


@orm.db_session
def set_alias(alias: str, alias_type: model.AliasType, **kwargs) -> model.PathAlias:
    """ Set a path alias.

    Arguments:

    alias -- The alias specification
    entry -- The entry to alias it to
    category -- The category to alias it to
    url -- The external URL to alias it to
    """

    spec = alias.split()
    path = spec[0]

    if not path or path[0] != '/':
        path = '/' + path

    values = {
        **kwargs,
        'path': urllib.parse.unquote(path),
        'alias_type': alias_type.value
    }

    if len(spec) > 1:
        values['template'] = spec[1]

    record = model.PathAlias.get(path=path)
    if record:
        record.set(**values)
    else:
        record = model.PathAlias(**values)

    orm.commit()
    return record


@orm.db_session
def remove_alias(path: str):
    """ Remove a path alias.

    Arguments:

    path -- the path to remove the alias of
    """
    orm.delete(p for p in model.PathAlias if p.path == path)  # type:ignore
    orm.commit()


@orm.db_session
def remove_aliases(target: typing.Union[model.Entry, model.Category]):
    """ Remove all aliases to a destination """

    if isinstance(target, model.Entry):
        orm.delete(p for p in model.PathAlias  # type:ignore
                   if p.entry == target)
    elif isinstance(target, model.Category):
        orm.delete(p for p in model.PathAlias  # type:ignore
                   if p.category == target)
    else:
        raise TypeError(f"Unknown type {type(target)}")
    orm.commit()


class Disposition:
    """ Alias disposition base class """
    # pylint:disable=too-few-public-methods

    def __init__(self):
        pass


class Response(Disposition):
    """ Disposition that provides a response directly """
    # pylint:disable=too-few-public-methods

    def __init__(self, response):
        super().__init__()
        self.response = response


class RenderEntry(Disposition):
    """ Disposition that requests an entry render """
    # pylint:disable=too-few-public-methods

    def __init__(self, entry: model.Entry, category: str, template: typing.Optional[str]):
        super().__init__()
        self.entry = entry
        self.category = category
        self.template = template


class RenderCategory(Disposition):
    """ Disposition that requests a category render """
    # pylint:disable=too-few-public-methods

    def __init__(self, category: str, template: typing.Optional[str]):
        super().__init__()
        self.category = category
        self.template = template


class AuthFailed(Disposition):
    """ Disposition that indicates that the pathalias was to an unauthorized entry """
    # pylint:disable=too-few-public-methods

    def __init__(self, cur_user: user.User, entry: model.Entry, category: str):
        super().__init__()
        self.cur_user = cur_user
        self.entry = entry
        self.category = category


def get_alias(path: str) -> typing.Optional[Disposition]:
    """ Get a path's alias mapping """
    from .flask_wrapper import current_app

    record = model.PathAlias.get(path=path)

    if not record or (record.entry and not record.entry.visible):
        url, permanent = current_app.test_path_regex(path)
        LOGGER.debug("regex match: %s %s", url, permanent)
        return Response(redirect(url, PERMANENT if permanent else TEMPORARY)) if url else None

    alias_type = model.AliasType(record.alias_type)

    category = (record.category.category if record.category
                else record.entry.category if record.entry
                else '')

    if record.entry:
        if record.entry.auth:
            LOGGER.debug("entry with auth")
            cur_user = user.get_active()
            if not record.entry.is_authorized(cur_user):
                return AuthFailed(cur_user, record.entry, category)

        if record.entry.redirect_url:
            LOGGER.debug("redirect URL: %s", record.entry.redirect_url)
            return Response(redirect(record.entry.redirect_url, PERMANENT))

    if alias_type == model.AliasType.MOUNT:
        LOGGER.debug("mount entry=%s category=%s template=%s",
                     record.entry, category, record.template)

        if record.entry:
            return RenderEntry(record.entry, category, record.template)

        return RenderCategory(category, record.template)

    LOGGER.debug("redirect entry=%s category=%s template=%s",
                 record.entry, category, record.template)
    args = {'category': category} if category else {}

    if record.template or not record.entry:
        endpoint = 'category'
        if record.template and record.template != 'index':
            args['template'] = record.template
        if record.entry:
            args['id'] = record.entry.id
    else:
        endpoint = 'entry'
        args['entry_id'] = record.entry.id
        if record.entry.slug_text:
            args['slug_text'] = record.entry.slug_text

    LOGGER.debug("endpoint=%s args=%s", endpoint, args)
    return Response(redirect(url_for(endpoint, **args), PERMANENT))



""" IndieAuth token endpoint """

import functools
import json
import logging
import time
import typing
import urllib.parse

import arrow
import flask
import itsdangerous
import requests
import werkzeug.exceptions as http_error
from pony import orm

from . import model, utils
from .config import config

LOGGER = logging.getLogger(__name__)


def signer(context: str):
    """ Gets the signer/validator for the tokens """
    from .flask_wrapper import current_app
    return itsdangerous.URLSafeSerializer(str(current_app.secret_key) + context)


def get_token(id_url: str, lifetime: int, scope: str = None, context: str = '') -> str:
    """ Gets a signed token for the given identity"""
    token = {'me': utils.canonicize_url(id_url)}
    if scope:
        token['scope'] = scope

    return signer(context).dumps((token, int(time.time() + lifetime)))


def parse_token(token: str, context: str = '') -> typing.Dict[str, str]:
    """ Parse a bearer token to get the stored data """
    try:
        ident, expires = signer(context).loads(token)
    except itsdangerous.BadData as error:
        LOGGER.error("Got token parse error: %s", error)
        flask.g.token_error = 'Invalid token'  # pylint:disable=assigning-non-slot
        raise http_error.Unauthorized('Invalid token') from error

    if expires < time.time():
        LOGGER.info("Got expired token for %s", ident['me'])
        flask.g.token_error = "Token expired"  # pylint:disable=assigning-non-slot
        raise http_error.Unauthorized("Token expired")

    return ident


def request(user):
    """ Called whenever an authenticated access fails; marks authentication
    as being upgradeable.

    Currently this is unused by Publ itself, but a site can make use of it to
    e.g. add a ``WWW-Authenticate`` header or the like in a post-request hook.
    """

    if not user:
        flask.g.needs_auth = True  # pylint:disable=assigning-non-slot


def send_auth_ticket(subject: str,
                     resource: str,
                     endpoint: str,
                     scope: str = None):
    """ Initiate the TicketAuth flow """
    from .flask_wrapper import current_app

    def _submit():
        ticket = get_token(subject, config.ticket_lifetime, scope, context='ticket')

        req = requests.post(endpoint, data={
            'ticket': ticket,
            'resource': resource,
            'subject': subject
        })
        LOGGER.info("Auth ticket sent to %s for %s: %d %s",
                    endpoint, subject, req.status_code, req.text)

    # Use the indexer's threadpool to issue the ticket in the background
    current_app.indexer.submit(_submit)


@orm.db_session()
def log_grant(identity: str):
    """ Update the user table with the granted token """
    import authl.handlers.indieauth

    values = {
        'last_token': arrow.utcnow().datetime,
    }

    profile = authl.handlers.indieauth.get_profile(identity)
    if profile:
        values['profile'] = profile

    record = model.KnownUser.get(user=identity)
    if record:
        record.set(**values)
    else:
        record = model.KnownUser(user=identity,
                                 **values,
                                 last_seen=arrow.utcnow().datetime)


def redeem_grant(grant_type: str, auth_token: str):
    """ Redeem a grant from a provided redemption ticket """
    grant = parse_token(auth_token, grant_type)
    LOGGER.info("Redeeming %s for %s; scopes=%s", grant_type, grant['me'],
                grant.get('scope'))

    scope = grant.get('scope', '')

    token = get_token(grant['me'], config.token_lifetime, scope)

    response = {
        'access_token': token,
        'token_type': 'Bearer',
        'me': grant['me'],
        'expires_in': config.token_lifetime,
        'refresh_token': get_token(grant['me'],
                                   config.refresh_token_lifetime,
                                   scope,
                                   context='refresh_token')
    }
    if scope:
        response['scope'] = scope

    log_grant(grant['me'])

    return json.dumps(response), {'Content-Type': 'application/json'}


@functools.lru_cache()
def get_ticket_endpoint(me_url: str):
    """ Get the IndieAuth Ticket Auth endpoint and the canonical identity URL """
    LOGGER.debug("get_ticket_endpoint %s", me_url)
    import authl.handlers.indieauth
    from bs4 import BeautifulSoup

    req = authl.utils.request_url(me_url)
    content = BeautifulSoup(req.text, 'html.parser')

    if req.links and 'canonical' in req.links:
        canonical_url = req.links['canonical']['url']
    else:
        link = content.find('link', rel='canonical')
        if link:
            canonical_url = urllib.parse.urljoin(me_url, link.get('href'))
        else:
            canonical_url = me_url

    if utils.canonicize_url(canonical_url) != utils.canonicize_url(me_url):
        # We have a rel="canonical" which mismatches the provided identity URL
        LOGGER.debug("%s -> canonical=%s", me_url, canonical_url)
        endpoint, me_url = authl.handlers.indieauth.find_endpoint(canonical_url,
                                                                  rel='ticket_endpoint')
    else:
        # Use our fetch to seed Authl's endpoint fetch and get that instead
        endpoints, me_url = authl.handlers.indieauth.find_endpoints(me_url,
                                                                    req.links, content)
        endpoint = endpoints.get('ticket_endpoint')

    LOGGER.debug("%s %s", me_url, endpoint)
    return endpoint, me_url


def ticket_request(me_url: str, scope: str):
    """ Initiate a ticket request """

    try:
        endpoint, me_url = get_ticket_endpoint(utils.canonicize_url(me_url))
    except RuntimeError:
        endpoint = None
    if not endpoint:
        raise http_error.BadRequest("Could not get ticket endpoint")
    LOGGER.info("endpoint: %s", endpoint)
    send_auth_ticket(me_url, flask.request.url_root, endpoint, scope)
    return "Ticket sent", 202


def parse_authorization_header(header):
    """ Parse an Authorization: header from an HTTP request into token """
    parts = header.split()
    if len(parts) < 2:
        raise http_error.BadRequest("Malformed authorization header")

    if parts[0].lower() == 'bearer':
        token = parse_token(parts[1])
        return token

    raise http_error.BadRequest(f"Unknown authorization type '{parts[0]}'")


def indieauth_endpoint():
    """ IndieAuth token endpoint """

    if 'grant_type' in flask.request.form:
        # token grant
        if flask.request.form['grant_type'] == 'ticket':
            # TicketAuth
            if 'ticket' not in flask.request.form:
                raise http_error.BadRequest("Missing ticket")
            return redeem_grant('ticket', flask.request.form['ticket'])

        if flask.request.form['grant_type'] == 'refresh_token':
            # Refresh token redemption
            if 'refresh_token' not in flask.request.form:
                raise http_error.BadRequest("Missing refresh_token")
            return redeem_grant('refresh_token', flask.request.form['refresh_token'])

        raise http_error.BadRequest("Unknown grant type")

    if 'action' in flask.request.form:
        # provisional ticket request flow, per https://github.com/indieweb/indieauth/issues/87
        if flask.request.form['action'] == 'ticket' and 'subject' in flask.request.form:
            return ticket_request(flask.request.form['subject'],
                                  flask.request.form.get('scope', ''))

        raise http_error.BadRequest()

    if 'Authorization' in flask.request.headers:
        # ticket verification
        token = parse_authorization_header(flask.request.headers['Authorization'])
        return json.dumps(token), {'Content-Type': 'application/json'}

    if 'me' in flask.request.args:
        # ad-hoc ticket request
        return ticket_request(flask.request.args['me'],
                              flask.request.args.get('scope', ''))

    raise http_error.BadRequest()



""" Base class for managed images """

import html
import typing
from abc import ABC, abstractmethod

import flask

from .. import utils

ImgSize = typing.Tuple[int, int]
RenditionAttrs = typing.Dict[str, typing.Any]


class Image(ABC):
    """ Base class for image handlers """

    def __init__(self, search_path):
        self.search_path = search_path

    @abstractmethod
    def _key(self):
        pass

    def __repr__(self):
        return repr((self._key(), repr(self.search_path)))

    def __hash__(self):
        return hash((self._key(), repr(self.search_path)))

    @abstractmethod
    def get_rendition(self, output_scale: int = 1, **kwargs) -> typing.Tuple[str, ImgSize]:
        """ Get a rendition of the image with the specified output scale and specification.

        Returns: a tuple of (url, size) for the image. """

    @abstractmethod
    def _get_img_attrs(self,
                       spec: utils.ArgDict,
                       style_parts: typing.List[str]) -> utils.TagAttrs:
        """ Get the img attributes for the given rendition arguments. Style parts
        should be appended to the style_parts instead.
        """

    @property
    @abstractmethod
    def _filename(self) -> str:
        """ Get the filename of the file, for default alt text purposes """

    def get_img_attrs(self, params: dict) -> utils.TagAttrs:
        """ Get an attribute list (src, style, et al) for the image.

        Returns: a dict of attributes e.g. {'src':'foo.jpg','srcset':'foo.jpg 1x, bar.jpg 2x']
        """

        params = utils.prefix_normalize(params)

        styles: typing.List[str] = []
        attrs = self._get_img_attrs(params, styles)

        attrs['loading'] = params.get('image_loading', 'lazy')

        for key in ('img_style', 'style'):
            styles += utils.as_list(params.get(key))

        shape = self._get_shape_style(params)
        if shape:
            styles.append(f"shape-outside: url('{shape}')")

        def set_val(key, val):
            if val is not None:
                attrs[key] = val

        if 'img_class' in params or 'class' in params:
            set_val('class', params.get('img_class', params.get('class')))
        if 'img_id' in params or 'id' in params:
            set_val('id', params.get('img_id', params.get('id')))
        if styles:
            set_val('style', ';'.join(styles))

        if '_mark_rewritten' in params:
            set_val('data-publ-rewritten', params['_mark_rewritten'])

        return attrs

    def get_img_tag(self,
                    title: str = None,
                    alt_text: str = None,
                    _show_thumbnail: bool = True,
                    **kwargs) -> str:
        """ Build a <img> tag for the image with the specified options.

        Returns: an HTML fragment. """

        try:
            attrs = self.get_img_attrs(kwargs)

            if alt_text:
                attrs['alt'] = alt_text
            else:
                attrs['alt'] = self._filename

            if title:
                attrs['title'] = title

            if _show_thumbnail:
                thumb = utils.make_tag(
                    'img', attrs, start_end=kwargs.get('xhtml', False))
            else:
                thumb = ''

            return flask.Markup(
                self._wrap_link_target(
                    kwargs,
                    thumb,
                    title))
        except FileNotFoundError as error:
            text = '<span class="error">Image not found: '
            text += f'<code>{html.escape(error.filename)}</code>'
            if ' ' in error.filename:
                text += ' (Did you forget a <code>|</code>?)'
            text += '</span>'
            return flask.Markup(text)

    def _get_shape_style(self, kwargs):
        shape = kwargs.get('shape')
        if not shape:
            return None

        # Only pull in size-related attributes (e.g. no format, background,
        # etc.)
        size_args = {k: v for k, v in kwargs.items() if k in (
            'width', 'height', 'max_width', 'max_height', 'absolute', 'crop'
        )}

        if shape is True:
            # if the shape is True, just return the base rendition
            url, _ = self.get_rendition(1, **size_args)
        else:
            # otherwise, the usual rules apply
            from . import get_image  # pylint: disable=cyclic-import
            other_image = get_image(shape, self.search_path)
            url, _ = other_image.get_rendition(1, **size_args)
        return url

    def get_css_background(self, uncomment: bool = False, **kwargs):
        """ Get the CSS background attributes for an element.

        Additional arguments:

        uncomment -- surround the attributes with `*/` and `/*` so that the
            template tag can be kept inside a comment block, to keep syntax
            highlighters happy
        """

        text = self._css_background(**kwargs)
        if uncomment:
            text = f' */ {text} /* '

        return text

    @abstractmethod
    def _css_background(self, **kwargs):
        """ Build CSS background-image properties that apply this image.

        Returns: a CSS fragment. """

    def __call__(self, *args, **kwargs):
        url, _ = self.get_rendition(*args, **kwargs)
        return url

    def __str__(self):
        return self()

    def _wrap_link_target(self, kwargs,
                          text: str,
                          title: typing.Optional[str]) -> str:
        if kwargs.get('link') is not None:
            if not text or kwargs['link'] is False:
                return text

            # pylint:disable=consider-using-f-string
            return '{}{}</a>'.format(
                utils.make_tag(
                    'a', {
                        'href': utils.remap_link_target(
                            kwargs['link'],
                            kwargs.get('absolute')),
                        'class': kwargs.get('link_class', False)
                    }),
                text)

        if kwargs.get('gallery_id'):
            return f'{self._fullsize_link_tag(kwargs, title)}{text}</a>'

        return text

    def get_fullsize(self, kwargs) -> str:
        """ Get the fullsize rendition URL """
        fullsize_args = {}

        if 'absolute' in kwargs:
            fullsize_args['absolute'] = kwargs['absolute']

        for key in ('width', 'height', 'quality', 'format', 'background', 'crop'):
            fsk = 'fullsize_' + key
            if fsk in kwargs:
                fullsize_args[key] = kwargs[fsk]

        img_fullsize, _ = self.get_rendition(1, **fullsize_args)
        return img_fullsize

    def _fullsize_link_tag(self, kwargs, title: typing.Optional[str]) -> str:
        """ Render an <a href> that points to the fullsize rendition specified """

        return utils.make_tag('a', {
            'href': self.get_fullsize(kwargs),
            'data-lightbox': kwargs.get('gallery_id', False),
            'title': title or False,
            'class': kwargs.get('link_class', False)
        })



# model.py
""" Database schema for the content index """

# pylint: disable=too-few-public-methods

import datetime
import logging
from enum import Enum

from pony import orm

db = orm.Database()  # pylint: disable=invalid-name

DbEntity: orm.core.Entity = db.Entity

LOGGER = logging.getLogger(__name__)

# schema version; bump this number if it changes
SCHEMA_VERSION = 22


class GlobalConfig(DbEntity):
    """ Global configuration data """
    key = orm.PrimaryKey(str)
    int_value = orm.Optional(int)


class PublishStatus(Enum):
    """ The status of the entry """
    DRAFT = 0       # Entry should not be rendered

    HIDDEN = 1      # Entry should be shown via direct link, but not shown on a view
    UNLISTED = 1    # Synonym for HIDDEN

    PUBLISHED = 2   # Entry is visible

    SCHEDULED = 3   # Entry will be visible in the future

    GONE = 4        # Entry is gone, won't be coming back
    DELETED = 4     # synonym for GONE

    ILLEGAL = 5     # taken down for legal reasons (error code 451)
    BURNT = 5
    BURNED = 5
    DMCA = 5

    TEAPOT = 6      # RFC 2324


class AliasType(Enum):
    """ The type of PathAlias mapping """
    REDIRECT = 0    # Redirect to the canon URL
    MOUNT = 1       # Display the URL at the alias


class FileFingerprint(DbEntity):
    """ File modification time """
    file_path = orm.PrimaryKey(str)
    fingerprint = orm.Required(str)
    file_mtime = orm.Required(float)


class Entry(DbEntity):
    """ Indexed entry """
    file_path = orm.Required(str, unique=True)
    category = orm.Optional(str)
    status = orm.Required(int)

    # UNIX epoch, for ordering and visibility
    utc_timestamp = orm.Required(int, size=64)

    # arbitrary timezone, for pagination
    local_date = orm.Required(datetime.datetime)

    # The actual displayable date - stored as string to guarantee the timezone
    # is maintained
    display_date = orm.Required(str)

    slug_text = orm.Optional(str)
    entry_type = orm.Optional(str)
    redirect_url = orm.Optional(str)   # maps to Redirect-To

    title = orm.Optional(str)
    sort_title = orm.Optional(str)

    aliases = orm.Set("PathAlias")
    tags = orm.Set("EntryTagged")

    auth = orm.Set("EntryAuth")
    auth_log = orm.Set("AuthLog")

    attachments = orm.Set("Entry", reverse="attached")
    attached = orm.Set("Entry", reverse="attachments")

    canonical_path = orm.Optional(str)

    orm.composite_index(category, entry_type, utc_timestamp)
    orm.composite_index(category, entry_type, local_date)
    orm.composite_index(category, entry_type, sort_title)

    @property
    def visible(self) -> bool:
        """ Returns true if the entry should be viewable """
        return self.status in (PublishStatus.HIDDEN.value,
                               PublishStatus.PUBLISHED.value,
                               PublishStatus.SCHEDULED.value)

    def is_authorized(self, user) -> bool:
        """ Returns whether the entry is visible to the specified user """
        if not self.auth:
            return True

        LOGGER.debug("Computing auth for entry %s user %s", self.file_path, user)

        if user and user.is_admin:
            LOGGER.debug("User is admin")
            return True

        result = False
        for auth in self.auth.order_by(EntryAuth.order):
            LOGGER.debug("  %d test=%s allowed=%s", auth.order, auth.user_group, auth.allowed)
            if auth.user_group == '*':
                # Special group * refers to all logged-in users
                result = auth.allowed == (user is not None)
                LOGGER.debug("  result->%s", result)

            else:
                if user and auth.user_group in user.auth_groups:
                    result = auth.allowed
                    LOGGER.debug("  result->%s", result)

        LOGGER.debug("Final result: %s", result)
        return result


class EntryTag(DbEntity):
    """ Tags available for entries """
    key = orm.PrimaryKey(str)
    name = orm.Required(str)

    entries = orm.Set("EntryTagged")


class EntryTagged(DbEntity):
    """ Actual tag membership for entries """
    entry = orm.Required(Entry)
    tag = orm.Required(EntryTag)
    hidden = orm.Required(bool)

    orm.composite_key(entry, tag)
    orm.composite_key(tag, entry)


class Category(DbEntity):
    """ Metadata for a category """

    category = orm.Optional(str, unique=True)
    file_path = orm.Required(str)
    sort_name = orm.Optional(str)

    aliases = orm.Set("PathAlias")


class PathAlias(DbEntity):
    """ Path alias mapping """
    path = orm.PrimaryKey(str)
    entry = orm.Optional(Entry)
    category = orm.Optional(Category)
    template = orm.Optional(str)
    alias_type = orm.Required(int)  # reflects AliasType


class Image(DbEntity):
    """ Image metadata """
    file_path = orm.PrimaryKey(str)
    checksum = orm.Required(str)
    fingerprint = orm.Required(str)

    width = orm.Optional(int)
    height = orm.Optional(int)
    transparent = orm.Optional(bool)

    is_asset = orm.Required(bool, default=False)
    asset_name = orm.Optional(str, index=True)


class EntryAuth(DbEntity):
    """ An authentication record for an entry """
    order = orm.Required(int)
    entry = orm.Required(Entry)
    user_group = orm.Required(str)
    allowed = orm.Required(bool)

    orm.composite_key(entry, order)


class KnownUser(DbEntity):
    """ Users who are known to the system """
    user = orm.PrimaryKey(str)
    last_login = orm.Optional(datetime.datetime)
    last_seen = orm.Required(datetime.datetime, index=True)
    last_token = orm.Optional(datetime.datetime)
    profile = orm.Optional(orm.Json)


class AuthLog(DbEntity):
    """ Authentication log for private entries """
    date = orm.Required(datetime.datetime, index=True)
    entry = orm.Required(Entry, index=True)
    user = orm.Required(str, index=True)
    user_groups = orm.Optional(str)  # group membership at the time of access
    authorized = orm.Required(bool, index=True)

    orm.PrimaryKey(entry, user)


def reset():
    """ Completely reset the database """
    LOGGER.info("Rebuilding schema")
    try:
        db.drop_all_tables(with_all_data=True)
        db.create_tables()
    except Exception as error:  # pylint:disable=broad-except
        raise RuntimeError("Unable to upgrade schema automatically; please " +
                           "delete the existing database and try again.") from error


def setup(config):
    """ Set up the database """
    rebuild = False

    try:
        db.bind(**config)
    except OSError:
        # Attempted to connect to a file-based database where the file didn't
        # exist
        db.bind(**config, create_db=True)

    try:
        db.generate_mapping(create_tables=True)
        with orm.db_session:
            version = GlobalConfig.get(key='schema_version')
            if not version or version.int_value != SCHEMA_VERSION:
                LOGGER.info("Wanted schema %d, got %s", SCHEMA_VERSION, version)
                rebuild = True
            else:
                LOGGER.info("Existing database has schema version %d",
                            version.int_value)

    except Exception:  # pylint:disable=broad-except
        LOGGER.exception("Error mapping schema")
        rebuild = True

    if rebuild:
        reset()

    with orm.db_session:
        if not GlobalConfig.get(key='schema_version'):
            LOGGER.info("setting schema version to %d", SCHEMA_VERSION)
            GlobalConfig(key='schema_version',
                         int_value=SCHEMA_VERSION)
            orm.commit()



""" Functions for handling content items """

import email
import hashlib
import logging
import os
import typing
import uuid

import arrow
import flask
import slugify
from pony import orm
from werkzeug.utils import cached_property

from . import (caching, cards, html_entry, links, markdown, model, path_alias,
               queries, tokens, user, utils)
from .config import config
from .utils import CallableProxy, CallableValue, TrueCallableProxy

LOGGER = logging.getLogger(__name__)


def load_message(filepath) -> email.message.Message:
    """ Load a message from the filesystem """
    with open(filepath, 'r', encoding='utf-8') as file:
        return email.message_from_file(file)


class Entry(caching.Memoizable):
    """ A wrapper for an entry. Lazily loads the actual message data when
    necessary.
    """

    # pylint: disable=too-many-instance-attributes,too-many-public-methods

    __hash__ = caching.Memoizable.__hash__  # type:ignore

    @staticmethod
    @utils.stash
    def load(record: model.Entry):
        """ Get a pooled Entry wrapper

        record -- the index record to use as the basis
        """
        return Entry(Entry.load.__name__, record)

    def __init__(self, create_key, record):
        """ Instantiate the Entry wrapper """

        assert create_key == Entry.load.__name__, "Entry must be created with Entry.load()"

        self._record = record   # index record
        self._fingerprint = model.FileFingerprint.get(file_path=record.file_path)
        LOGGER.debug('loaded entry %d, fingerprint=%s', record.id,
                     self._fingerprint.fingerprint if self._fingerprint else None)

        # maps (section,footnotes_enabled) -> toc/footnote counter
        self._counters: typing.Dict[typing.Tuple[str, bool], markdown.ItemCounter] = {}

    def __lt__(self, other):
        # pylint:disable=protected-access
        return self._record.id < other._record.id

    def _key(self):
        return self._record.id, self._record.file_path, self._fingerprint.fingerprint

    @cached_property
    def date(self) -> arrow.Arrow:
        """ Get the display date of the entry, as an Arrow object """
        return arrow.get(self._record.display_date)

    @cached_property
    def date_year(self) -> str:
        """ Get the entry date' year, useful for grouping """
        return self.date.format(utils.YEAR_FORMAT)

    @cached_property
    def date_month(self) -> str:
        """ Get the entry date's month, useful for grouping """
        return self.date.format(utils.MONTH_FORMAT)

    @cached_property
    def date_day(self) -> str:
        """ Get the entry date's day, useful for grouping """
        return self.date.format(utils.DAY_FORMAT)

    @cached_property
    def link(self) -> typing.Callable[..., str]:
        """ Get a link to this entry. Accepts the same parameters as permalink;
        may be pre-redirected. """
        def _link(*args, **kwargs) -> str:
            """ Returns a link, potentially pre-redirected """
            if self._record.redirect_url:
                return links.resolve(self._record.redirect_url,
                                     self.search_path, kwargs.get('absolute', False))

            return self.permalink(*args, **kwargs)

        return CallableProxy(_link)

    @cached_property
    def permalink(self) -> typing.Callable[..., str]:
        """ Get a canonical link to this entry. Accepts the following parameters:

        absolute -- if True, return an absolute/portable link (default: False)
        expand -- if True, expands the link to include the category and slug text;
            if False, it will only be the entry ID (default: True)
        """
        def _permalink(absolute=False, expand=True, **kwargs) -> str:
            if not self.authorized:
                expand = False

            if self._record.canonical_path:
                # This is a hack that assumes that the standard '/<template>'
                # rule is in effect. This will have to change if we implement
                # https://github.com/PlaidWeb/Publ/issues/286
                return flask.url_for('category',
                                     template=self._record.canonical_path,
                                     _external=absolute)

            return flask.url_for('entry',
                                 entry_id=self._record.id,
                                 category=self._record.category if expand else None,
                                 slug_text=self._record.slug_text
                                 if expand and self._record.slug_text
                                 else None,
                                 _external=absolute,
                                 **kwargs)

        return CallableProxy(_permalink)

    @cached_property
    def login(self) -> typing.Callable[..., str]:
        """ Get a link specifically for logging in to the entry. Not intended for general use;
        might be useful for some future authentication flow. """
        def _loginlink(absolute=False, **kwargs) -> str:
            pagelink = flask.url_for('entry', entry_id=self._record.id, **kwargs)
            return flask.url_for('login', redir=pagelink[1:], _external=absolute)
        return CallableProxy(_loginlink)

    @cached_property
    def private(self) -> bool:
        """ Returns True if this post is private, i.e. it is invisible to the logged-out user """
        return not self._record.is_authorized(None)

    @cached_property
    def archive(self) -> typing.Callable[..., str]:
        """ Get a link to this entry in the context of a category template.
        Accepts the following arguments:

        paging -- Which pagination scheme to use; one of:
            day -- the entry date's day
            month -- the entry date's month
            year -- the entry date's year
            offset -- count-based pagination starting at the entry (default)
        tag -- which tag(s) to filter on
        category -- Which category to generate the link against (default: the entry's category)
        template -- Which template to generate the link for
        """
        def _archive_link(paging=None, template='', category=None, absolute=False, tag=None) -> str:
            # pylint:disable=too-many-arguments
            args = {
                'template': template,
                'category': category if category is not None else self.category,
            }
            if paging == 'day':
                args['date'] = self.date.format(utils.DAY_FORMAT)
            elif paging == 'month':
                args['date'] = self.date.format(utils.MONTH_FORMAT)
            elif paging == 'year':
                args['date'] = self.date.format(utils.YEAR_FORMAT)
            elif paging == 'week':
                args['date'] = self.date.span('week')[0].format(utils.WEEK_FORMAT)
            elif paging == 'offset' or not paging:
                args['id'] = self._record.id
            else:
                raise ValueError(f"Unknown paging type '{paging}'")

            if tag:
                args['tag'] = list(utils.TagSet(utils.as_list(tag)).keys())

            return flask.url_for('category', **args, _external=absolute)

        return CallableProxy(_archive_link)

    @cached_property
    def type(self) -> str:
        """ An alias for entry_type """
        return self.entry_type

    @cached_property
    def tags(self) -> typing.List[str]:
        """ Get the original (non-normalized, non-folded) tags for the entry """
        return self.get_all('Tag')

    @cached_property
    def status(self) -> model.PublishStatus:
        """ Returns a typed version of the entry status """
        return model.PublishStatus(self.status)

    @cached_property
    def next(self) -> typing.Callable[..., typing.Optional["Entry"]]:
        """ Get the next entry in the category, ordered by date.

        Accepts view parameters as arguments.
        """
        def _next(**kwargs) -> typing.Optional["Entry"]:
            """ Get the next item in any particular category """
            spec = self._pagination_default_spec(kwargs)
            spec.update(kwargs)

            query = queries.build_query(spec)
            query = queries.where_after_entry(query, self._record)

            cur_user = user.get_active()
            for record in query.order_by(model.Entry.local_date,
                                         model.Entry.id):
                if record.is_authorized(cur_user):
                    return Entry.load(record)

                LOGGER.debug("User unauthorized for entry %d", record.id)
                tokens.request(cur_user)
            return None
        return CallableProxy(_next)

    @cached_property
    def previous(self) -> typing.Callable[..., typing.Optional["Entry"]]:
        """ Get the previous entry in the category, ordered by date.

        Accepts view parameters as arguments.
        """
        def _previous(**kwargs) -> typing.Optional["Entry"]:
            """ Get the previous item in any particular category """
            spec = self._pagination_default_spec(kwargs)
            spec.update(kwargs)

            query = queries.build_query(spec)
            query = queries.where_before_entry(query, self._record)

            cur_user = user.get_active()
            for record in query.order_by(orm.desc(model.Entry.local_date),
                                         orm.desc(model.Entry.id)):
                if record.is_authorized(cur_user):
                    return Entry.load(record)

                LOGGER.debug("User unauthorized for entry %d", record.id)
                tokens.request(cur_user)
            return None
        return CallableProxy(_previous)

    @cached_property
    def category(self):
        """ Get the category this entry belongs to. """
        from .category import Category  # pylint: disable=cyclic-import
        return Category.load(self._record.category)

    @cached_property
    def title(self) -> typing.Callable[..., str]:
        """ Get the title of the entry. Accepts the following arguments:

        markup -- If True, convert it from Markdown to HTML; otherwise, strip
            all markup (default: True)
        no_smartquotes -- if True, preserve quotes and other characters as originally
            presented
        markdown_extensions -- a list of markdown extensions to use
        always_show -- always show the title even if the current user is not
            authorized to see the entry
        """
        def _title(markup=True, markdown_extensions=None,
                   always_show=False, **kwargs) -> str:
            if not always_show and not self.authorized:
                return ''
            smartquotes = kwargs.get('smartquotes', not kwargs.get('no_smartquotes', False))
            return markdown.render_title(self._record.title, markup, smartquotes,
                                         markdown_extensions)
        return TrueCallableProxy(_title) if self._record.title else CallableValue('')

    @cached_property
    def search_path(self) -> typing.Tuple[str, ...]:
        """ The relative image search path for this entry """
        return (os.path.dirname(self._record.file_path), self.category.search_path)

    @cached_property
    def _message(self) -> email.message.Message:
        """ get the message payload """
        LOGGER.debug("Loading entry %d from %s", self._record.id, self._record.file_path)
        filepath = self._record.file_path
        try:
            return load_message(filepath)
        except FileNotFoundError:
            expire_record(self._record)
            empty = email.message.Message()
            empty.set_payload('')
            return empty

    @cached_property
    def _entry_content(self) -> typing.Tuple[str, str, bool]:
        if not self.authorized:
            return '', '', False

        body, _, more = self._message.get_payload().partition('\n.....\n')
        if not more and body.startswith('.....\n'):
            # The entry began with a cut, which failed to parse.
            # This rule is easier/faster than dealing with a regex from
            # hell.
            more = body[6:]
            body = ''

        _, ext = os.path.splitext(self._record.file_path)
        is_markdown = ext == '.md'

        return body, more, is_markdown

    @cached_property
    def body(self) -> typing.Callable[..., str]:
        """ Get the above-the-fold entry body text """
        body, _, is_markdown = self._entry_content

        def _body(**kwargs) -> str:
            LOGGER.debug("Rendering body; args=%s", kwargs)

            footnotes: typing.List[str] = []
            tocs: markdown.TocBuffer = []

            counter = markdown.ItemCounter()
            body_text = self._get_markup(body, is_markdown, args=kwargs,
                                         footnote_buffer=footnotes,
                                         toc_buffer=tocs,
                                         counter=counter)

            self._set_counter('body', kwargs, counter)

            return body_text

        return TrueCallableProxy(_body) if body else CallableValue('')

    @cached_property
    def more(self) -> typing.Callable[..., str]:
        """ Get the below-the-fold entry body text """
        _, more, is_markdown = self._entry_content

        def _more(**kwargs) -> str:
            LOGGER.debug("Rendering more; kwargs=%s", kwargs)

            body_count = self._get_counter('body', kwargs)
            LOGGER.debug("intro footnotes=%d tocs=%d codeblocks=%d",
                         body_count.footnote, body_count.toc, body_count.code_blocks)

            counter = body_count.copy()
            more_text = self._get_markup(more, is_markdown,
                                         args=kwargs,
                                         counter=counter)

            self._set_counter('more', kwargs, counter)

            return more_text

        return TrueCallableProxy(_more) if more else CallableValue('')

    @cached_property
    def footnotes(self) -> typing.Callable[..., str]:
        """ Get the rendered footnotes for the entry """
        body, more, is_markdown = self._entry_content

        def _footnotes(**kwargs) -> str:
            return self._get_footnotes(body, more, kwargs)

        if is_markdown:
            body_count = self._counters.get(('body', True))
            more_count = self._counters.get(('more', True))

            if ((body_count and body_count.footnote)
                    or (more_count and more_count.footnote)):
                return TrueCallableProxy(_footnotes)
            if body_count is None or more_count is None:
                return CallableProxy(_footnotes)

        return CallableValue('')

    @cached_property
    def toc(self) -> typing.Callable[..., str]:
        """ Get the rendered table of contents for the entry """
        body, more, is_markdown = self._entry_content

        def _toc(max_depth=None, **kwargs) -> str:
            return self._get_toc(body, more, max_depth, kwargs)

        if is_markdown:
            body_count = self._counters.get(('body', True))
            more_count = self._counters.get(('more', True))

            if ((body_count and body_count.toc)
                    or (more_count and more_count.toc)):
                return TrueCallableProxy(_toc)

            if body_count is None or more_count is None:
                return CallableProxy(_toc)

        return CallableValue('')

    @cached_property
    def card(self) -> typing.Callable[..., str]:
        """ Get the entry's OpenGraph card """

        def _get_card(**kwargs) -> str:
            """ Render out the tags for a Twitter/OpenGraph card for this entry. """

            def og_tag(key, val) -> str:
                """ produce an OpenGraph tag with the given key and value """
                return utils.make_tag('meta', {'property': key, 'content': val}, start_end=True)

            tags = og_tag('og:title', self.title(markup=False))
            tags += og_tag('og:url', self.link(absolute=True))

            body, more, is_markdown = self._entry_content
            html_text = self._get_markup(body + '\n\n' + more,
                                         is_markdown,
                                         args={'count': 1,
                                               **kwargs,
                                               "max_scale": 1,
                                               "_suppress_footnotes": True,
                                               "_no_resize_external": True,
                                               "absolute": True},
                                         counter=markdown.ItemCounter())
            card = cards.extract_card(html_text)

            for (image, width, height) in card.images[:kwargs.get('count', 1)]:
                tags += og_tag('og:image', image)
                if width:
                    tags += og_tag('og:image:width', width)
                if height:
                    tags += og_tag('og:image:height', height)

            description = self.summary(markup=False)
            if description:
                tags += og_tag('og:description', description)

            return flask.Markup(tags)

        return CallableProxy(_get_card)

    @cached_property
    def summary(self) -> typing.Callable[..., str]:
        """ Get the summary of the entry, falling back to the first paragraph if
        not present. Accepts the following arguments:

        markup -- If True, convert it from Markdown to HTML; otherwise, strip
            all markup (default: True)
        no_smartquotes -- if True, preserve quotes and other characters as originally
            presented
        markdown_extensions -- a list of markdown extensions to use
        always_show -- always show the title even if the current user is not
            authorized to see the entry
        """
        def _summary(markup=True, markdown_extensions=None,
                     always_show=False, **kwargs) -> str:
            if not always_show and not self.authorized:
                return ''

            summary = self.get('Summary')
            if summary:
                smartquotes = kwargs.get('smartquotes', not kwargs.get('no_smartquotes', False))
                return markdown.render_title(summary, markup, smartquotes,
                                             markdown_extensions)

            # We don't have a declared summary, so derive it from the first text paragraph
            body, more, is_markdown = self._entry_content

            if body or more:
                html_text = self._get_markup(body or more,
                                             is_markdown,
                                             args={**kwargs,
                                                   "max_scale": 1,
                                                   "_suppress_images": True,
                                                   "_suppress_footnotes": True,
                                                   "_no_resize_external": True,
                                                   "absolute": True},
                                             counter=markdown.ItemCounter())
                html_text = html_entry.first_paragraph(html_text)
                if markup:
                    return flask.Markup(html_text)

                return html_entry.strip_html(
                    html_text,
                    remove_elements=markdown.PLAINTEXT_REMOVE_ELEMENTS)

            return ''

        return CallableProxy(_summary)

    @cached_property
    def last_modified(self) -> arrow.Arrow:
        """ Get the date of last file modification """
        if self.get('Last-Modified'):
            return arrow.get(self.get('Last-Modified'))
        return self.date

    @property
    def authorized(self) -> bool:
        """ Returns if the current user is authorized to see this entry """
        return self._record.is_authorized(user.get_active())

    def _get_markup(self, text, is_markdown, args,
                    counter: markdown.ItemCounter,
                    footnote_buffer: typing.Optional[list] = None,
                    toc_buffer: typing.Optional[markdown.TocBuffer] = None,
                    postprocess: bool = True) -> str:
        """ get the rendered markup for an entry

            is_markdown -- whether the entry is formatted as Markdown
            kwargs -- parameters to pass to the Markdown processor
        """
        # pylint:disable=too-many-arguments
        if is_markdown:
            # Set defaults for the ID link generators, so permalinks from category
            # pages work correctly
            default_link = self.link(absolute=args.get('absolute'))
            for link_flag in ('footnotes_link', 'toc_link', 'code_number_links'):
                if link_flag not in args or args[link_flag] is True:
                    args[link_flag] = default_link

            return markdown.to_html(
                text,
                args=args,
                search_path=self.search_path,
                entry_id=self._record.id,
                footnote_buffer=footnote_buffer,
                toc_buffer=toc_buffer,
                postprocess=postprocess,
                counter=counter
            )

        text = html_entry.process(
            text,
            args,
            search_path=self.search_path)

        if not args.get('markup', True):
            text = html_entry.strip_html(text)

        return text

    @cached_property
    def attachments(self) -> typing.Callable[..., typing.List]:
        """ Returns a view of entries that are attached to this one. Takes the
        standard view arguments. """

        def _get_attachments(order=None, **kwargs) -> typing.List:
            query = queries.build_query({**kwargs,
                                         'attachments': self._record
                                         })
            if order:
                query = query.order_by(*queries.ORDER_BY[order])
            cur_user = user.get_active()

            return [Entry.load(e) for e in query
                    if e.is_authorized(cur_user) or tokens.request(cur_user)]

        return CallableProxy(_get_attachments)

    @cached_property
    def attached(self) -> typing.Callable[..., typing.List]:
        """ Get all the entries that have attached this one """

        def _get_attached(order=None, **kwargs) -> typing.List:
            query = queries.build_query({**kwargs,
                                         'attached': self._record
                                         })
            if order:
                query = query.order_by(*queries.ORDER_BY[order])
            return [Entry.load(e) for e in query]

        return CallableProxy(_get_attached)

    def _get_footnotes(self, body, more, args) -> str:
        """ get the rendered Markdown footnotes for the entry """
        footnotes: typing.List[str] = []
        counter = markdown.ItemCounter()
        if body and self._get_counter('body', args).footnote:
            self._get_markup(body, True, args=args,
                             footnote_buffer=footnotes,
                             postprocess=False,
                             counter=counter)
        if more and self._get_counter('more', args).footnote:
            self._get_markup(more, True, args=args,
                             footnote_buffer=footnotes,
                             counter=counter)

        if footnotes:
            return flask.Markup(f"<ol>{''.join(footnotes)}</ol>")
        return ''

    def _get_toc(self, body, more, max_depth, args) -> str:
        """ get the rendered ToC for the entry """
        tocs: markdown.TocBuffer = []
        args = {**args, '_suppress_footnotes': True}
        counter = markdown.ItemCounter()
        if body and self._get_counter('body', args).toc:
            self._get_markup(body, True, args=args, toc_buffer=tocs, postprocess=False,
                             counter=counter)
        if more and self._get_counter('more', args).toc:
            self._get_markup(more, True, args=args, toc_buffer=tocs, counter=counter)

        if tocs:
            return flask.Markup(markdown.toc_to_html(tocs, max_depth))
        return ''

    def _get_counter(self, section, args) -> markdown.ItemCounter:
        """ Count the countables given the specified section and arguments """
        body, more, is_markdown = self._entry_content
        if not is_markdown:
            return markdown.ItemCounter()

        footnotes = 'footnotes' in args.get('markdown_extensions', config.markdown_extensions)

        if (section, footnotes) in self._counters:
            return self._counters[(section, footnotes)]

        if section == 'body':
            text = body
        elif section == 'more':
            text = more
        else:
            raise ValueError("Unknown content section " + section)

        if text:
            counter = markdown.get_counters(text, args)
            self._counters[(section, footnotes)] = counter
            return counter

        return markdown.ItemCounter()

    def _set_counter(self, section, args, counter: markdown.ItemCounter):
        """ Register the counts that we already know """
        footnotes = 'footnotes' in args.get('markdown_extensions', config.markdown_extensions)
        self._counters[(section, footnotes)] = counter

    def _authorized_attr(self, name):
        """ Return whether an attribute is authorized to be read """
        return name.lower() in ('uuid', 'id', 'date', 'last-modified') or self.authorized

    def __getattr__(self, name):
        """ Proxy undefined properties to the backing objects """

        # Only allow a few vital things for unauthorized access
        if not self._authorized_attr(name):
            return None

        # Don't pass certain things through the database
        if name.lower() not in ('auth') and hasattr(self._record, name):
            return getattr(self._record, name)

        return self._message.get(name)

    def _pagination_default_spec(self, kwargs):
        category = kwargs.get('category', self._record.category)
        return {
            'category': category,
        }

    def get(self, name, default=None, always_show=False) -> typing.Optional[str]:
        """ Get a single header on an entry """
        if always_show or self._authorized_attr(name):
            return self._message.get(name, default)
        return None

    def get_all(self, name, always_show=False) -> typing.List[str]:
        """ Get all related headers on an entry, as an iterable list """
        if always_show or self._authorized_attr(name):
            values = self._message.get_all(name)
            return [str(item) for item in values] if values else []
        return []

    def __eq__(self, other) -> bool:
        if isinstance(other, int):
            return other == self._record.id
        # pylint:disable=protected-access
        return isinstance(other, Entry) and (other is self or other._record == self._record)

    @staticmethod
    def filter_auth(entries, count=None, unauthorized=0):
        """ Filter a list of entries based on authorization, with a maximum
        unauthorized entry count """

        result: typing.List[Entry] = []
        cur_user = user.get_active()
        for record in entries:
            if count is not None and len(result) >= count:
                break

            auth = record.is_authorized(cur_user)
            if auth or unauthorized:
                result.append(Entry.load(record))
                if not auth and unauthorized is not True:
                    unauthorized -= 1

            if not auth:
                tokens.request(cur_user)

        return result


def get_entry_id(entry, fullpath, assign_id) -> typing.Optional[int]:
    """ Get or generate an entry ID for an entry """
    other_entry: typing.Optional[model.Entry] = None

    try:
        entry_id = int(entry['Entry-ID']) if 'Entry-ID' in entry else None
    except (ValueError, KeyError, TypeError) as err:
        LOGGER.debug("Invalid entry-id: %s", err)

    # See if we've inadvertently duplicated an entry ID
    if entry_id is not None:
        try:
            other_entry = model.Entry.get(id=entry_id)
            if (other_entry
                    and os.path.isfile(other_entry.file_path)
                    and not os.path.samefile(other_entry.file_path, fullpath)
                    and other_entry.status != model.PublishStatus.DRAFT.value):
                entry_id = None
            else:
                other_entry = None
        except FileNotFoundError:
            # the other file doesn't exist, so just let it go
            pass

    # Do we need to assign a new ID?
    if not entry_id and not assign_id:
        # We're not assigning IDs yet
        return None

    if not entry_id:
        # See if we already have an entry with this file path
        by_filepath = model.Entry.select(lambda e: e.file_path == fullpath).first()
        if by_filepath:
            entry_id = by_filepath.id

    if not entry_id:
        # We still don't have an ID; generate one pseudo-randomly, based on the
        # entry file path. This approach averages around 0.25 collisions per ID
        # generated while keeping the entry ID reasonably short. In general,
        # count*N averages 1/(N-1) collisions per ID.

        limit = max(10, orm.get(orm.count(e)
                                for e in model.Entry) * 5)  # type:ignore
        attempt = 0
        while not entry_id or model.Entry.get(id=entry_id):
            # Stably generate a quasi-random entry ID from the file path
            md5 = hashlib.md5()
            md5.update(f"{fullpath} {attempt}".encode('utf-8'))
            entry_id = int.from_bytes(md5.digest(), byteorder='big') % limit
            attempt = attempt + 1

    if other_entry:
        LOGGER.warning("Entry '%s' had ID %d, which belongs to '%s'. Reassigned to %d",
                       fullpath, other_entry.id, other_entry.file_path, entry_id)

    return entry_id


def save_file(fullpath: str, entry: email.message.Message, fingerprint: str):
    """ Save a message file out, without mangling the headers """
    from atomicwrites import atomic_write
    with atomic_write(fullpath, overwrite=True) as file:
        # we can't just use file.write(str(entry)) because otherwise the
        # headers "helpfully" do MIME encoding normalization.
        # str(val) is necessary to get around email.header's encoding
        # shenanigans
        for key, val in entry.items():
            print(f'{key}: {str(val)}', file=file)
        print('', file=file)
        file.write(entry.get_payload())

        if utils.file_fingerprint(fullpath) != fingerprint:
            LOGGER.warning("File %s changed during atomic write; aborting", fullpath)
            raise RuntimeError("File changed during reindex")

    return True


@orm.db_session(retry=5)
def scan_file(fullpath: str, relpath: typing.Optional[str], fixup_pass: int) -> bool:
    """ scan a file and put it into the index

    :param fullpath str: The full file path
    :param relpath typing.Optional[str]: The file path relative to the content
        root; if None, this will be inferred
    :param fixup_pass int: Which iteration of fixing-up we're on

    """
    # pylint: disable=too-many-branches,too-many-statements,too-many-locals
    from .flask_wrapper import current_app

    try:
        check_fingerprint = utils.file_fingerprint(fullpath)
        entry = load_message(fullpath)
    except FileNotFoundError:
        # The file doesn't exist, so remove it from the index
        record = model.Entry.get(file_path=fullpath)
        if record:
            expire_record(record)
        return True

    entry_id = get_entry_id(entry, fullpath, fixup_pass > 0)
    if entry_id is None:
        return False

    fixup_needed = False

    if not relpath:
        relpath = os.path.relpath(fullpath, config.content_folder)

    title = entry.get('title', '')

    values = {
        'file_path': fullpath,
        'category': entry.get('Category', utils.get_category(relpath)),
        'status': model.PublishStatus[entry.get('Status', 'SCHEDULED').upper()].value,
        'entry_type': entry.get('Entry-Type', ''),
        'slug_text': slugify.slugify(
            entry.get('Slug-Text',
                      markdown.render_title(title, markup=False, smartquotes=False))),
        'redirect_url': entry.get('Redirect-To', ''),
        'title': title,
        'sort_title': entry.get('Sort-Title', title),
        'canonical_path': entry.get('Path-Canonical', '')
    }

    entry_date = None
    if 'Date' in entry:
        try:
            entry_date = arrow.get(entry['Date'], tzinfo=config.timezone)
        except arrow.parser.ParserError:
            entry_date = None
    if entry_date is None:
        del entry['Date']
        entry_date = arrow.get(
            os.stat(fullpath).st_ctime).to(config.timezone)
        entry['Date'] = entry_date.format()
        fixup_needed = True

    if 'Last-Modified' in entry:
        last_modified_str = entry['Last-Modified']
        try:
            last_modified = arrow.get(
                last_modified_str, tzinfo=config.timezone)
        except arrow.parser.ParserError:
            last_modified = arrow.get()
            del entry['Last-Modified']
            entry['Last-Modified'] = last_modified.format()
            fixup_needed = True

    values['display_date'] = entry_date.isoformat()
    values['utc_timestamp'] = entry_date.to('utc').int_timestamp
    values['local_date'] = entry_date.naive

    LOGGER.debug("getting entry %s with id %d", fullpath, entry_id)

    remove_by_path(fullpath, entry_id)

    record = model.Entry.get(id=entry_id)
    if record:
        LOGGER.debug("Reusing existing entry %d", record.id)
        record.set(**values)
    else:
        record = model.Entry(id=entry_id, **values)

    # Update the entry ID
    if str(record.id) != entry['Entry-ID']:  # pylint:disable=no-member
        del entry['Entry-ID']
        entry['Entry-ID'] = str(record.id)
        fixup_needed = True

    if 'UUID' not in entry:
        entry['UUID'] = str(uuid.uuid5(
            uuid.NAMESPACE_URL, 'file://' + fullpath))
        fixup_needed = True

    # add other relationships to the index
    path_alias.remove_aliases(record)
    if record.visible:
        for alias in entry.get_all('Path-Alias', []):
            path_alias.set_alias(alias, model.AliasType.REDIRECT, entry=record)
        for alias in entry.get_all('Path-Mount', []):
            path_alias.set_alias(alias, model.AliasType.MOUNT, entry=record)
        for alias in entry.get_all('Path-Canonical', []):
            path_alias.set_alias(alias, model.AliasType.MOUNT, entry=record)

    orm.delete(p for p in model.EntryAuth if p.entry == record)  # type:ignore
    orm.commit()
    for order, user_group in enumerate(entry.get('Auth', '').split()):
        allowed = (user_group[0] != '!')
        if not allowed:
            user_group = user_group[1:]
        model.EntryAuth(order=order, entry=record, user_group=user_group, allowed=allowed)
    orm.commit()

    with orm.db_session:
        set_tags = {
            utils.tag_key(t[0]): t
            for t in [(k, True) for k in entry.get_all('Hidden-Tag', [])]
            + [(k, False) for k in entry.get_all('Tag', [])]
        }
        remove_tags = []

        for etag in record.tags:
            LOGGER.debug("  has tag %s,%s", etag.tag.key, etag.tag.name)
            if etag.tag.key not in set_tags:
                remove_tags.append(etag)
        LOGGER.debug("set_tags %s remove_tags %s", set_tags, remove_tags)

        for etag in remove_tags:
            tag = etag.tag
            etag.delete()
            if len(tag.entries) == 0:
                LOGGER.debug("tag %s/%s entry count went to 0", tag.key, tag.name)
                tag.delete()

        for (key, tag) in set_tags.items():
            name, hidden = tag

            # get the underlying tag object
            tag_record = model.EntryTag.get(key=key)
            if not tag_record:
                LOGGER.debug("creating tag %s/%s", key, name)
                tag_record = model.EntryTag(key=key, name=name)
            elif name != tag_record.name and name != key and not hidden:
                LOGGER.debug("updating tag name %s/%s -> %s",
                             key, tag_record.name, name)
                tag_record.name = name

            # get the tag placement object
            etag = model.EntryTagged.get(tag=tag_record, entry=record)
            if not etag:
                etag = model.EntryTagged(tag=tag_record, entry=record, hidden=hidden)
            else:
                etag.hidden = hidden
            record.tags.add(etag)

        orm.commit()

    result = True

    # manage entry attachments
    with orm.db_session:
        from .category import search_path as cat_search_path
        search_path = (os.path.dirname(fullpath), cat_search_path(record.category))

        set_attach = set()
        for attach in entry.get_all('Attach', []):
            other = links.find_entry(attach, search_path)
            if other:
                set_attach.add(other)
            elif fixup_pass < 3:
                # The entry hasn't been found, so treat this as a fixup task
                # Pass 0 - this entry might not have an ID
                # Pass 1 - the other entry might not have an ID (since this can be scheduled
                #    before pass 1 of the other entry)
                # Pass 2 - everything should have an ID now
                LOGGER.info("Attempted to link to unknown entry '%s -> %s'; retrying",
                            relpath, attach)
                result = False
            else:
                LOGGER.warning(
                    "Failed to link to unknown entry '%s -> %s'; ignoring", relpath, attach)

        remove_attach = []
        for attach in record.attachments:
            if attach not in set_attach:
                remove_attach.append(attach)

        LOGGER.debug("set_attach %s remove_attach %s", set_attach, remove_attach)
        for attach in remove_attach:
            record.attachments.remove(attach)
        for attach in set_attach:
            record.attachments.add(attach)

        orm.commit()

    # do final fixups
    if record.status == model.PublishStatus.DRAFT.value:
        LOGGER.info("Not touching draft entry %s", fullpath)
    elif fixup_needed:
        LOGGER.info("Fixing up entry %s", fullpath)
        result = save_file(fullpath, entry, check_fingerprint)

    # register with the search index
    current_app.search_index.update(record, entry)

    return result


@orm.db_session
def expire_record(record):
    """ Expire a record for a missing entry """

    # This entry no longer exists so delete anything that relies on it
    orm.delete(pa for pa in model.PathAlias if pa.entry == record)

    # mark the entry as GONE to remove it from indexes
    record.status = model.PublishStatus.GONE.value
    orm.commit()


@orm.db_session
def remove_by_path(fullpath: str, entry_id: int):
    """ Remove entries for a path that don't match the expected ID """

    orm.delete(pa for pa in model.PathAlias  # type:ignore
               if pa.entry.file_path == fullpath
               and pa.entry.id != entry_id)
    orm.delete(e for e in model.Entry  # type:ignore
               if e.file_path == fullpath
               and e.id != entry_id)
    orm.commit()



# markdown.py
""" markdown formatting functionality """
# pylint:disable=consider-using-f-string

import html
import logging
import re
import typing
import urllib.parse

import flask
import misaka
import pygments
import pygments.formatters
import pygments.lexers
import slugify

from . import html_entry, image, links, utils
from .config import config

LOGGER = logging.getLogger(__name__)

TocEntry = typing.Tuple[int, str]
TocBuffer = typing.List[TocEntry]

# Allow these tags in TOC entries
TOC_ALLOWED_TAGS = ('sup', 'sub',
                    'em', 'strong',
                    'b', 'i',
                    'code',
                    'del', 'add', 'mark')

# Remove these tags from plaintext-style conversions
PLAINTEXT_REMOVE_ELEMENTS = ('del', 's', 'table')


class ItemCounter(misaka.BaseRenderer):
    """ Just counts the number of things in an entry without any further processing """

    def __init__(self, toc: int = 0, footnote: int = 0, code_blocks: int = 0):
        super().__init__()
        self.toc = toc
        self.footnote = footnote
        self.code_blocks = code_blocks

    def __str__(self):
        return 'ItemCounter(footnote={footnote},toc={toc},code_blocks={bc})'.format(
            footnote=self.footnote,
            toc=self.toc,
            bc=self.code_blocks
        )

    def copy(self):
        """ Return a copy of this counter """
        return ItemCounter(self.toc, self.footnote, self.code_blocks)

    def header(self, content, level):
        """ count this header """
        # pylint:disable=unused-argument
        self.toc += 1

    def footnote_def(self, content, num):
        """ count this footnote """
        # pylint:disable=unused-argument
        self.footnote += 1

    def blockcode(self, text, lang):
        """ count this blockcode """
        # pylint:disable=unused-argument
        self.code_blocks += 1


class HtmlCodeFormatter(pygments.formatters.HtmlFormatter):  # pylint:disable=no-member
    """ Customized code block formatter

    Constructor arguments:

    line_id_prefix -- the prefix for each line's <span> id
    link_base -- the base URL for the line links
    """
    # pylint:disable=too-few-public-methods

    def __init__(self,
                 line_id_prefix: str,
                 link_base: typing.Union[str, bool],
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.line_id_prefix = line_id_prefix
        self.link_base = link_base

    def wrap(self, source, outfile):
        """ called by pygments """
        # pylint:disable=unused-argument
        return self._wrap_code(source)

    def _wrap_code(self, source):
        line_number = 0
        for i, line in source:
            if i == 1:
                line_number += 1
                line_id = f"{self.line_id_prefix}L{line_number}"

                yield 1, (utils.make_tag('span', {'class': 'line',
                                                  'id': line_id})
                          + (utils.make_tag('a', {
                              'class': 'line-number',
                              'href': self.link_base + '#' + line_id})
                             + '</a>' if self.link_base else '')
                          + utils.make_tag('span', {'class': 'line-content'})
                          + line.rstrip('\n')
                          + '</span></span>\n')


class HtmlRenderer(misaka.HtmlRenderer):
    """ Customized renderer for enhanced Markdown formatting

    :param dict config: The configuration for the Markdown tags
    :param list search_path: Directories to look in for resolving relatively-linked files
    :param int entry_id: the numeric entry ID
    :param list footnote_buffer: the buffer of footnote entries so far
    :param list toc_buffer: the buffer of TOC entries so far
    :param ItemCounter counter: item count accumulator

    For configuration values, see the individual Markdown implementation functions,
    :py:func:`HtmlRenderer.image`, and py:mod:`publ.image`.

    """

    def __init__(self, args: typing.Dict,
                 search_path: typing.Tuple[str],
                 entry_id: typing.Optional[int],
                 footnote_buffer: typing.Optional[typing.List[str]],
                 toc_buffer: typing.Optional[TocBuffer],
                 counter: ItemCounter):
        # pylint:disable=no-member,too-many-arguments
        super().__init__(0, args.get('xhtml') and misaka.HTML_USE_XHTML or 0)

        self._config = args
        self._search_path = search_path
        self._entry_id = entry_id

        self._footnote_ofs = counter.footnote
        self._footnote_buffer = footnote_buffer

        self._toc_buffer = toc_buffer

        self._counter = counter

    def _inner(self, text):
        """ process some inner markdown """
        return misaka.Markdown(self,
                               self._config.get('markdown_extensions',
                                                config.markdown_extensions))(text)

    @staticmethod
    def footnotes(_):
        """ Stub function; footnote rendering is handled by
        :py:func:entry.entry.footnotes """
        return None

    def _footnote_num(self, num):
        return num + self._footnote_ofs

    def _footnote_id(self, num, anchor):
        return f'{anchor}_e{self._entry_id}_fn{self._footnote_num(num)}'

    def _footnote_url(self, num, anchor):
        return urllib.parse.urljoin(self._config.get('footnotes_link', ''),
                                    '#' + self._footnote_id(num, anchor))

    def footnote_ref(self, num):
        """ Render a link to a footnote

        Relevant configuration options:
        * ``footnotes_class``: The class to apply to a footnote marker
        * ``footnotes_link``: The base URL for footnote links
        """

        if self._config.get('_suppress_footnotes'):
            return '\u200b'  # zero-width space to prevent Misaka fallback

        return '{sup}{link}{content}</a></sup>'.format(
            sup=utils.make_tag('sup', {
                'id': self._footnote_id(num, "r"),
                'class': self._config.get('footnotes_class', False)
            }),
            link=utils.make_tag('a', {
                'href': self._footnote_url(num, "d"),
                'rel': 'footnote'
            }),
            content=self._footnote_num(num))

    def footnote_def(self, content, num):
        """ Renders the body of a footnote to be used by :py:func:`entry.Entry.footnotes`.

        Relevant configuration options:

        * ``footnotes_link``: The base URL for footnote links
        * ``footnotes_return``: The return symbol on an expanded footnote
          (default: ``'↩'``)
        """

        LOGGER.debug("footnote_def %d: %s", num, content)

        self._counter.footnote_def(content, num)

        # Insert the return anchor before the end of the first content block
        before, partition, after = content.partition('</p>')
        text = '{li}{before}&nbsp;{link}{icon}</a>{partition}{after}</li>'.format(
            li=utils.make_tag('li', {
                'id': self._footnote_id(num, "d")
            }),
            before=before,
            link=utils.make_tag('a', {
                'href': self._footnote_url(num, "r"),
                'rev': 'footnote'
            }),
            icon=self._config.get('footnotes_return', '↩'),
            partition=partition,
            after=after,
        )

        self._footnote_buffer.append(text)

    def _header_id(self, content, level, num):
        """ Return a reasonable anchor ID for a heading."""
        return '{eid}_h{level}_{num}_{slug}'.format(
            eid=self._entry_id,
            level=level,
            num=num,
            slug=slugify.slugify(content, max_length=32))

    def header(self, content, level):
        """ Make a header with anchor.

        Relevant configuration:

        * ``toc_link``: The base URL for table of contents links from a heading
        * ``heading_link_class``: The class to apply to a TOC link from a heading
        * ``heading_link_attrs``: Additional attributes to apply to TOC links from headings
        * ``heading_template``: The format template for a heading's TOC link
          (default: ``{link}</a>{{text}}``). Receives the following template arguments:
            * ``{link}``: The ``<a>`` tag linking to the TOC entry
            * ``{text}``: The text of the heading
         """

        self._counter.header(content, level)
        htag = f'h{level}'

        hid = self._header_id(html_entry.strip_html(content,
                                                    remove_elements=PLAINTEXT_REMOVE_ELEMENTS),
                              level, self._counter.toc)

        atag = utils.make_tag('a', {
            'href': urllib.parse.urljoin(self._config.get('toc_link', ''),
                                         '#' + hid),
            'class': self._config.get('heading_link_class', False),
            **self._config.get('heading_link_attrs', {})
        })

        if self._toc_buffer is not None:
            LOGGER.debug("append toc: %d %s", level, content)
            self._toc_buffer.append(
                (level,
                 '{atag}{content}</a>'.format(
                     atag=atag,
                     content=html_entry.strip_html(content, allowed_tags=TOC_ALLOWED_TAGS))))

        content = self._config.get('heading_template', '{link}</a>{text}').format(
            link=atag,
            text=content)

        return '{htag_open}{content}</{htag}>'.format(
            htag_open=utils.make_tag(htag, {'id': hid}),
            content=content,
            htag=htag)

    def image(self, raw_url, title='', alt=''):
        """ Adapt a standard Markdown image to a generated rendition set.

        Extended syntax::

            ![alt text{container_args}](image1.jpg{img_args} "title text"
            |image2.jpg{img_args} "title text"
            |...)

        Configuration for each image comes from the page template, then from the
        ``container_args``, then from the ``img_args``.

        Configuration that applies to the image set itself:

        * ``figure``: If set, wrap the image set in a ``<figure>`` tag. If this
          value is a string, use the string value as the CSS class name.

        * ``caption``: If set, add the text as a ``<figcaption>``. Markdown supported.
          Implies ``figure=True``.

        * ``div_class``: The CSS class name to use on any wrapper div

        * ``div_style``: Additional CSS styles to apply to the wrapper div

        * ``count``: The maximum number of images to show at once

        * ``more_text``: If there are more than ``count`` images, add this text
          indicating that there are more images to be seen. This string gets two template
          arguments, ``{count}`` which is the total number of images in the set, and
          ``{remain}`` which is the number of images omitted from the set.

        * ``more_link``: If `more_text` is shown, this will format the text as a link
          to this location.

        * ``more_class``: If `more_text` is shown, applies this CSS class to the link

        See :py:mod:`publ.image` for configuration for image renditions.

        """
        # pylint: disable=too-many-locals

        # If we aren't generating images or markup, there's no reason to do any of this
        if self._config.get('_suppress_images') or not self._config.get('markup', True):
            return ' '

        text = ''

        image_specs = raw_url
        if title:
            image_specs += f' "{title}"'

        alt, container_args = image.parse_img_config(alt)

        container_args = utils.prefix_normalize({**self._config, **container_args})

        spec_list = image.get_spec_list(image_specs, container_args)

        remain = 0
        for (spec, show) in spec_list:
            text += self._render_image(spec, show,
                                       container_args,
                                       alt)
            if not show:
                remain += 1

        if remain and 'more_text' in container_args:
            more_text = container_args['more_text'].format(
                count=len(spec_list),
                remain=remain)
            if 'more_link' in container_args:
                more_text = '{a}{text}</a>'.format(
                    text=more_text,
                    a=utils.make_tag('a', {
                        'href': container_args['more_link'],
                        'class': container_args.get('more_class', False)
                    }))
            text += flask.Markup(more_text)

        if text and (container_args.get('div_class') or
                     container_args.get('div_style')):
            text = '{tag}{text}</div>'.format(
                tag=utils.make_tag('div',
                                   {'class': container_args.get('div_class') or False,
                                    'style': container_args.get('div_style') or False}),
                text=text)

        if text and (container_args.get('figure') or container_args.get('caption')):
            fig_class = container_args.get('figure')
            if container_args.get('caption'):
                caption = '<figcaption>' + utils.strip_single_paragraph(
                    self._inner(container_args['caption'])) + '</figcaption>'
            else:
                caption = ''
            text = '{tag}{text}{caption}</figure>'.format(
                tag=utils.make_tag('figure',
                                   {'class': fig_class if isinstance(fig_class, str) else False}),
                text=text,
                caption=caption)

        # if text is ''/falsy then misaka interprets this as a failed parse...
        return text or ' '

    def blockcode(self, text, lang):
        """ Pass a code fence through pygments, and add line-numbering scaffolding.

        Relevant configuration:

        * ``code_highlight``: Whether to apply syntax highlighting to fenced
          code blocks (default: ``True``)
        * ``code_number_links``: Whether to add line-numbering links to fenced
          code blocks' lines. If set to a string, specifies the base URL for
          these links (default: ``True``)

        """
        LOGGER.debug("blockcode lang=%s", lang)

        self._counter.blockcode(text, lang)

        out = '<figure class="blockcode">'

        lang, _, args = utils.parse_spec(lang, 0)
        if args:
            args = {**self._config, **args}
        else:
            args = self._config
        LOGGER.debug("Code: language=%s args=%s", lang, args)

        if text.startswith('!'):
            caption, _, text = text.partition('\n')
            caption = self._inner(caption[1:]).strip()
            if caption:
                out += '<figcaption>' + utils.strip_single_paragraph(caption) + '</figcaption>'
        elif text.startswith(r'\!') or text.startswith('\n'):
            # Allow the initial ! to be escaped out, or skip an initial blank line
            text = text[1:]

        highlight = lang and args.get('code_highlight', True)
        number_lines = highlight and args.get('code_number_links', True)

        if highlight:
            try:
                lexer = pygments.lexers.get_lexer_by_name(lang or 'text', stripall=True)
            except pygments.lexers.ClassNotFound:
                lexer = pygments.lexers.TextLexer()  # pylint:disable=no-member
        else:
            lexer = None

        out += utils.make_tag('pre', {
            'class': 'highlight' if highlight else False,
            'data-language': lang if lang else False,
            'data-line-numbers': None if number_lines else False,
        } if lang else {})

        if lexer:
            link_base = number_lines
            if link_base is True:
                # It was set explicitly True, which means we need to re-inherit the
                # default from the template
                link_base = self._config.get('code_number_links')
            formatter = HtmlCodeFormatter(
                line_id_prefix=f"e{self._entry_id}cb{self._counter.code_blocks}",
                link_base=link_base,
            )
            out += pygments.highlight(str(text), lexer, formatter)
        else:
            lines = text.splitlines()
            for line in lines:
                out += ('<span class="line"><span class="line-content">'
                        + html.escape(line) + '</span></span>\n')

        out += '</pre></figure>'

        return out

    def link(self, content, link, title=''):
        """ Emit a link, potentially remapped based on our embed or static rules.

        Relevant configuration:

        * ``absolute``: Whether to produce absolute/external, rather than relative,
          links (default: ``False``)
        """

        if not self._config.get('markup', True):
            return content

        link = links.resolve(link, self._search_path,
                             self._config.get('absolute'))

        return '{}{}</a>'.format(
            utils.make_tag('a', {
                'href': link,
                'title': title if title else False
            }),
            content)

    @staticmethod
    def paragraph(content):
        """ emit a paragraph """

        # if the content contains a top-level block element then don't wrap it in a <p>
        # tag
        for element in ('div', 'figure'):
            if content.startswith(f'<{element}') and content.endswith(f'</{element}>'):
                return '\n' + content + '\n'

        return '<p>' + content + '</p>'

    def _render_image(self, spec, show, container_args, alt_text=None):
        """ Render an image specification into an <img> tag """

        try:
            path, image_args, title = image.parse_image_spec(spec)
        except Exception as err:  # pylint: disable=broad-except
            LOGGER.exception("Got error on spec %s: %s", spec, err)
            return flask.Markup('<span class="error">Couldn\'t parse image spec: ' +
                                '<code>{}</code> {}</span>'.format(flask.escape(spec),
                                                                   flask.escape(str(err))))

        composite_args = {**container_args, **image_args}

        try:
            img = image.get_image(path, self._search_path)
            return img.get_img_tag(title,
                                   alt_text,
                                   **composite_args,
                                   _show_thumbnail=show,
                                   _mark_rewritten=True)
        except Exception as err:  # pylint: disable=broad-except
            LOGGER.exception("Got error on image %s: %s", path, err)
            return flask.Markup('<span class="error">Error loading image {}: {}</span>'.format(
                flask.escape(spec), flask.escape(str(err))))


def to_html(text, args, search_path,
            counter: ItemCounter,
            entry_id: typing.Optional[int] = None,
            toc_buffer: typing.Optional[TocBuffer] = None,
            footnote_buffer: typing.Optional[typing.List[str]] = None,
            postprocess: bool = True):
    """ Convert Markdown text to HTML.

    toc_buffer -- a list of (level,text) for all headings in the entry

    footnote_buffer -- a list that will contain <li>s with the footnote items, if
    there are any footnotes to be found.

    postprocess -- whether to postprocess the buffers for smartypants/HTML/etc.
    """
    # pylint:disable=too-many-arguments

    LOGGER.debug("counter: %s %s", id(counter), counter)

    footnotes: typing.List[str] = footnote_buffer if footnote_buffer is not None else []
    tocs: TocBuffer = toc_buffer if toc_buffer is not None else []

    # first process as Markdown
    renderer = HtmlRenderer({**config.layout, **args},
                            search_path,
                            toc_buffer=tocs,
                            footnote_buffer=footnotes,
                            entry_id=entry_id,
                            counter=counter)
    processor = misaka.Markdown(renderer,
                                args.get('markdown_extensions', config.markdown_extensions))
    text = processor(text)

    if postprocess:
        # convert smartquotes, if so configured.
        # We prefer setting 'smartquotes' but we fall back to the negation of
        # 'no_smartquotes' for backwards compatibility with a not-well-considered
        # API.
        if 'no_smartquotes' in args:
            LOGGER.warning("no_smartquotes is deprecated and will be removed in a future version")
        smartquotes = args.get('smartquotes', not args.get('no_smartquotes', False))
        if smartquotes:
            text = misaka.smartypants(text)
            footnotes[:] = (misaka.smartypants(text) for text in footnotes)
            tocs[:] = ((level, misaka.smartypants(text)) for level, text in tocs)

        # now filter through html_entry to rewrite local src/href links
        text = html_entry.process(text, args, search_path)
        footnotes[:] = (html_entry.process(text, args, search_path)
                        for text in footnotes)

    return flask.Markup(text)


def get_counters(text, args):
    """ Count the number of stateful items in Markdown text. """
    counter = ItemCounter()
    processor = misaka.Markdown(counter,
                                args.get('markdown_extensions', config.markdown_extensions))
    processor(text)
    return counter


class TitleRenderer(HtmlRenderer):
    """ A renderer that is suitable for rendering out page titles and nothing else """

    def __init__(self):
        super().__init__({}, [], entry_id=0, toc_buffer=[], footnote_buffer=[],
                         counter=ItemCounter())

    @staticmethod
    def paragraph(content):
        """ Passthrough """
        return content

    @staticmethod
    def list(content, is_ordered, is_block):
        """ Passthrough """
        # pylint: disable=unused-argument
        return content

    @staticmethod
    def listitem(content, is_ordered, is_block):
        """ Just add the * back on """
        # pylint: disable=unused-argument
        if not is_ordered:
            return '* ' + content
        raise ValueError("Not sure how we got here")

    @staticmethod
    def header(content, level):
        """ Passthrough """
        # pylint: disable=unused-argument
        return content


def render_title(text, markup=True, smartquotes=True, markdown_extensions=None):
    """ Convert a Markdown title to HTML """

    # If the title starts with something that looks like a list, save it for
    # later
    pfx, text = re.match(r'([0-9. ]*)(.*)', text).group(1, 2)

    text = pfx + misaka.Markdown(TitleRenderer(),
                                 extensions=markdown_extensions
                                 or config.markdown_extensions)(text)

    if smartquotes:
        text = misaka.smartypants(text)

    if markup:
        return flask.Markup(text)

    return html_entry.strip_html(text, remove_elements=PLAINTEXT_REMOVE_ELEMENTS)


def toc_to_html(toc: TocBuffer, max_level: int = None) -> str:
    """ Convert a TocBuffer to an appropriate <ol> """

    if not toc:
        return ''

    cur_level = 0
    out = ''

    # preprocess: find the lowest heading level
    min_level = min([level for level, _ in toc])
    toc = [(level - min_level + 1, text) for level, text in toc]

    for level, text in toc:
        if max_level is None or level <= max_level:
            if level > cur_level:
                # open sublists
                out += '<ol><li>' * (level - cur_level)
            else:
                # close sublists and the prior entry from this level
                out += '</li></ol>' * (cur_level - level) + '</li><li>'
            out += text
            cur_level = level

    out += '</li></ol>' * cur_level
    return out
